"""Experimental flag"""
import logging
import sys
from dataclasses import dataclass
from functools import wraps
from typing import Dict, List, Optional

import click

from samcli.cli.context import Context
from samcli.cli.global_config import ConfigEntry, GlobalConfig
from samcli.commands._utils.parameterized_option import parameterized_option
from samcli.lib.utils.colors import Colored

LOG = logging.getLogger(__name__)

EXPERIMENTAL_PROMPT = """
This feature is currently in beta.
Visit the docs page to learn more about the AWS Beta terms https://aws.amazon.com/service-terms/.
Enter Y to proceed with the command, or enter N to cancel:
"""

EXPERIMENTAL_WARNING = """
Experimental features are enabled for this session.
Visit the docs page to learn more about the AWS Beta terms https://aws.amazon.com/service-terms/.
"""

EXPERIMENTAL_ENV_VAR_PREFIX = "SAM_CLI_BETA_"


@dataclass(frozen=True, eq=True)
class ExperimentalEntry(ConfigEntry):
    """Child data class of ConfigEntry that enforces
    config_key and env_var_key to be not None"""

    config_key: str
    env_var_key: str
    persistent: bool = False


class ExperimentalFlag:
    """Class for storing all experimental related ConfigEntries"""

    All = ExperimentalEntry("experimentalAll", EXPERIMENTAL_ENV_VAR_PREFIX + "FEATURES")
    BuildPerformance = ExperimentalEntry(
        "experimentalBuildPerformance", EXPERIMENTAL_ENV_VAR_PREFIX + "BUILD_PERFORMANCE"
    )
    IaCsSupport = {
        "terraform": ExperimentalEntry(
            "experimentalTerraformSupport", EXPERIMENTAL_ENV_VAR_PREFIX + "TERRAFORM_SUPPORT"
        )
    }
    RustCargoLambda = ExperimentalEntry("experimentalCargoLambda", EXPERIMENTAL_ENV_VAR_PREFIX + "RUST_CARGO_LAMBDA")


def is_experimental_enabled(config_entry: ExperimentalEntry) -> bool:
    """Whether a given experimental flag is enabled or not.
    If experimentalAll is set to True, then it will always return True.

    Parameters
    ----------
    config_entry : ExperimentalEntry
        Experimental flag ExperimentalEntry

    Returns
    -------
    bool
        Whether the experimental flag is enabled or not.
    """
    gc = GlobalConfig()
    enabled = gc.get_value(config_entry, default=False, value_type=bool, is_flag=True)
    if not enabled:
        enabled = gc.get_value(ExperimentalFlag.All, default=False, value_type=bool, is_flag=True)
    return enabled


def set_experimental(config_entry: ExperimentalEntry = ExperimentalFlag.All, enabled: bool = True) -> None:
    """Set the experimental flag to enabled or disabled.

    Parameters
    ----------
    config_entry : ExperimentalEntry, optional
        Flag to be set, by default ExperimentalFlag.All
    enabled : bool, optional
        Enabled or disabled, by default True
    """
    gc = GlobalConfig()
    gc.set_value(config_entry, enabled, is_flag=True, flush=False)


def get_all_experimental() -> List[ExperimentalEntry]:
    """
    Returns
    -------
    List[ExperimentalEntry]
        List all experimental flags in the ExperimentalFlag class.
    """
    all_experimental_flags = []
    for name in dir(ExperimentalFlag):
        if name.startswith("__"):
            continue
        value = getattr(ExperimentalFlag, name)
        if isinstance(value, ExperimentalEntry):
            all_experimental_flags.append(value)
        elif isinstance(value, dict):
            all_experimental_flags += value.values()
    return all_experimental_flags


def get_all_experimental_statues() -> Dict[str, bool]:
    """Get statues of all experimental flags in a dictionary.

    Returns
    -------
    Dict[str, bool]
        Dictionary with key as configuration value and value as enabled or disabled.
    """
    return {entry.config_key: is_experimental_enabled(entry) for entry in get_all_experimental() if entry.config_key}


def get_enabled_experimental_flags() -> List[str]:
    """
    Returns a list of string, which contains enabled experimental flags for current session

    Returns
    -------
    List[str]
        List of strings which contains all enabled experimental flag names
    """
    enabled_experimentals = []
    for experimental_key, status in get_all_experimental_statues().items():
        if status:
            enabled_experimentals.append(experimental_key)
    return enabled_experimentals


def disable_all_experimental():
    """Turn off all experimental flags in the ExperimentalFlag class."""
    for entry in get_all_experimental():
        set_experimental(entry, False)


def update_experimental_context(show_warning=True):
    """Set experimental for the current click context.

    Parameters
    ----------
    show_warning : bool, optional
        Should warning be shown, by default True
    """
    if not Context.get_current_context().experimental:
        Context.get_current_context().experimental = True
        if show_warning:
            LOG.warning(Colored().yellow(EXPERIMENTAL_WARNING))


def _experimental_option_callback(ctx, param, enabled: Optional[bool]):
    """Click parameter callback for --beta-features or --no-beta-features.
    If neither is specified, enabled will be None.
    If --beta-features is set, enabled will be True,
    we should turn on all experimental flags.
    If --no-beta-features is set, enabled will be False,
    we should turn off all experimental flags, overriding existing env vars.
    """
    if enabled is None:
        return

    if enabled:
        set_experimental(ExperimentalFlag.All, True)
        update_experimental_context()
    else:
        disable_all_experimental()


def experimental_click_option(default: Optional[bool]):
    return click.option(
        "--beta-features/--no-beta-features",
        default=default,
        required=False,
        is_flag=True,
        expose_value=False,
        callback=_experimental_option_callback,
        help="Should beta features be enabled.",
    )


@parameterized_option
def experimental(f, default: Optional[bool] = None):
    """Decorator for adding --beta-features and --no-beta-features click options to a command."""
    return experimental_click_option(default)(f)


@parameterized_option
def force_experimental(
    f, config_entry: ExperimentalEntry = ExperimentalFlag.All, prompt=EXPERIMENTAL_PROMPT, default=None
):
    """Decorator for adding --beta-features and --no-beta-features click options to a command.
    If experimental flag env var or --beta-features flag is not specified, this will then
    prompt the user for confirmation.
    The program will exit if confirmation is denied.
    """

    def wrap(func):
        @wraps(func)
        def wrapped_func(*args, **kwargs):
            if not prompt_experimental(config_entry=config_entry, prompt=prompt):
                sys.exit(1)
            return func(*args, **kwargs)

        return wrapped_func

    return experimental_click_option(default)(wrap(f))


@parameterized_option
def force_experimental_option(
    f, option: str, config_entry: ExperimentalEntry = ExperimentalFlag.All, prompt=EXPERIMENTAL_PROMPT
):
    """Decorator for making a specific option to be experimental.
    A prompt will be shown if experimental is not enabled and the option is specified.
    """

    def wrap(func):
        @wraps(func)
        def wrapped_func(*args, **kwargs):
            if kwargs[option]:
                if not prompt_experimental(config_entry=config_entry, prompt=prompt):
                    sys.exit(1)
            return func(*args, **kwargs)

        return wrapped_func

    return wrap(f)


def prompt_experimental(
    config_entry: ExperimentalEntry = ExperimentalFlag.All, prompt: str = EXPERIMENTAL_PROMPT
) -> bool:
    """Prompt the user for experimental features.
    If the corresponding experimental flag is already specified, the prompt will be skipped.
    If confirmation is granted, the corresponding experimental flag env var will be set.

    Parameters
    ----------
    config_entry : ExperimentalEntry, optional
        Which experimental flag should be set, by default ExperimentalFlag.All
    prompt : str, optional
        Text to be shown in the prompt, by default EXPERIMENTAL_PROMPT

    Returns
    -------
    bool
        Whether user have accepted the experimental feature.
    """
    if is_experimental_enabled(config_entry):
        update_experimental_context()
        return True
    confirmed = click.confirm(Colored().yellow(prompt), default=False)
    if confirmed:
        set_experimental(config_entry=config_entry, enabled=True)
        update_experimental_context()
    return confirmed



"""SyncFlow base class """
import logging
from abc import ABC, abstractmethod
from enum import Enum
from os import environ
from pathlib import Path
from threading import Lock
from typing import TYPE_CHECKING, Any, Dict, List, NamedTuple, Optional, Set, cast

from boto3.session import Session

from samcli.lib.providers.provider import ResourceIdentifier, Stack, get_resource_by_id
from samcli.lib.sync.exceptions import MissingLockException, MissingPhysicalResourceError
from samcli.lib.utils.boto_utils import get_boto_client_provider_from_session_with_config
from samcli.lib.utils.lock_distributor import LockChain, LockDistributor

if TYPE_CHECKING:  # pragma: no cover
    from samcli.commands.build.build_context import BuildContext
    from samcli.commands.deploy.deploy_context import DeployContext
    from samcli.commands.sync.sync_context import SyncContext

# Logging with multiple processes is not safe. Use a log queue in the future.
# https://docs.python.org/3/howto/logging-cookbook.html#:~:text=Although%20logging%20is%20thread%2Dsafe,across%20multiple%20processes%20in%20Python.
LOG = logging.getLogger(__name__)


def get_default_retry_config() -> Optional[Dict]:
    """
    Returns a default retry config if nothing is overriden by environment variables
    """
    if environ.get("AWS_MAX_ATTEMPTS") or environ.get("AWS_RETRY_MODE"):
        return None
    return {"max_attempts": 10, "mode": "standard"}


class ApiCallTypes(Enum):
    """API call stages that can be locked on"""

    BUILD = "Build"
    UPDATE_FUNCTION_CONFIGURATION = "UpdateFunctionConfiguration"
    UPDATE_FUNCTION_CODE = "UpdateFunctionCode"


class ResourceAPICall(NamedTuple):
    """Named tuple for a resource and its potential API calls"""

    shared_resource: str
    api_calls: List[ApiCallTypes]


class SyncFlow(ABC):
    """Base class for a SyncFlow"""

    _log_name: str
    _build_context: "BuildContext"
    _deploy_context: "DeployContext"
    _sync_context: "SyncContext"
    _stacks: Optional[List[Stack]]
    _session: Optional[Session]
    _physical_id_mapping: Dict[str, str]
    _locks: Optional[Dict[str, Lock]]
    # Local hash represents the state of a particular sync flow
    # We store the hash value in sync state toml file as value
    _local_sha: Optional[str]

    def __init__(
        self,
        build_context: "BuildContext",
        deploy_context: "DeployContext",
        sync_context: "SyncContext",
        physical_id_mapping: Dict[str, str],
        log_name: str,
        stacks: Optional[List[Stack]] = None,
    ):
        """
        Parameters
        ----------
        build_context : BuildContext
            BuildContext used for build related parameters
        deploy_context : BuildContext
            DeployContext used for this deploy related parameters
        sync_context: SyncContext
            SyncContext object that obtains sync information.
        physical_id_mapping : Dict[str, str]
            Mapping between resource logical identifier and physical identifier
        log_name : str
            Name to be used for logging purposes
        stacks : List[Stack], optional
            List of stacks containing a root stack and optional nested stacks
        """
        self._build_context = build_context
        self._deploy_context = deploy_context
        self._sync_context = sync_context
        self._log_name = log_name
        self._stacks = stacks
        self._session = None
        self._physical_id_mapping = physical_id_mapping
        self._locks = None
        self._local_sha = None

    def set_up(self) -> None:
        """Clients and other expensives setups should be handled here instead of constructor"""
        pass

    def _get_session(self) -> Session:
        if not self._session:
            self._session = Session(profile_name=self._deploy_context.profile, region_name=self._deploy_context.region)
        return self._session

    def _boto_client(self, client_name: str):
        default_retry_config = get_default_retry_config()
        if not default_retry_config:
            LOG.debug("Creating boto client (%s) with user's retry config", client_name)
            return get_boto_client_provider_from_session_with_config(self._get_session())(client_name)

        LOG.debug("Creating boto client (%s) with default retry config", client_name)
        return get_boto_client_provider_from_session_with_config(self._get_session(), retries=default_retry_config)(
            client_name
        )

    @property
    @abstractmethod
    def sync_state_identifier(self) -> str:
        """
        Sync state is the unique identifier for each sync flow
        We store the identifier in sync state toml file as key
        """
        raise NotImplementedError("sync_state_identifier")

    @abstractmethod
    def gather_resources(self) -> None:
        """Local operations that need to be done before comparison and syncing with remote
        Ex: Building lambda functions
        """
        raise NotImplementedError("gather_resources")

    def _update_local_hash(self) -> None:
        """Updates the latest local hash of the sync flow which then can be used for comparison for next run"""
        if not self._local_sha:
            LOG.debug("%sNo local hash is configured, skipping to update local hash", self.log_prefix)
            return

        self._sync_context.update_resource_sync_state(self.sync_state_identifier, self._local_sha)

    def compare_local(self) -> bool:
        """Comparison between local resource and its local stored state.
        If the resources are identical, sync and gather dependencies will be skipped.
        Simply return False if there is no comparison needed.
        Ex: Comparing local Lambda function artifact with stored SHA256

        Returns
        -------
        bool
            Return True if current resource and cached are in sync. Skipping rest of the execution.
            Return False otherwise.
        """
        stored_sha = self._sync_context.get_resource_latest_sync_hash(self.sync_state_identifier)
        if self._local_sha and stored_sha and self._local_sha == stored_sha:
            pass
            # return True
        return False

    @abstractmethod
    def compare_remote(self) -> bool:
        """Comparison between local and remote resources.
        This can be used for optimization if comparison is a lot faster than sync.
        If the resources are identical, sync and gather dependencies will be skipped.
        Simply return False if there is no comparison needed.
        Ex: Comparing local Lambda function artifact with remote SHA256

        Returns
        -------
        bool
            Return True if local and remote are in sync. Skipping rest of the execution.
            Return False otherwise.
        """
        raise NotImplementedError("compare_remote")

    @abstractmethod
    def sync(self) -> None:
        """Step that syncs local resources with remote.
        Ex: Call UpdateFunctionCode for Lambda Functions
        """
        raise NotImplementedError("sync")

    @abstractmethod
    def gather_dependencies(self) -> List["SyncFlow"]:
        """Gather a list of SyncFlows that should be executed after the current change.
        This can be sync flows for other resources that depends on the current one.
        Ex: Update Lambda functions if a layer sync flow creates a new version.

        Returns
        ------
        List[SyncFlow]
            List of sync flows that need to be executed after the current one finishes.
        """
        raise NotImplementedError("update_dependencies")

    @abstractmethod
    def _get_resource_api_calls(self) -> List[ResourceAPICall]:
        """Get resources and their associating API calls. This is used for locking purposes.
        Returns
        -------
        Dict[str, List[str]]
            Key as resource logical ID
            Value as list of api calls that the resource can make
        """
        raise NotImplementedError("_get_resource_api_calls")

    def has_locks(self) -> bool:
        """Check if a sync flow has locks and needs to enter a lock context
        Returns
        -------
        bool
            whether or not a sync flow contains locks
        """
        return bool(self._locks)

    def get_lock_keys(self) -> Set[str]:
        """Get a list of function + API calls that can be used as keys for LockDistributor

        Returns
        -------
        Set[str]
            Set of keys for all resources and their API calls
        """
        lock_keys = set()
        for resource_api_calls in self._get_resource_api_calls():
            for api_call in resource_api_calls.api_calls:
                lock_keys.add(SyncFlow._get_lock_key(resource_api_calls.shared_resource, api_call))
        return lock_keys

    def set_locks_with_distributor(self, distributor: LockDistributor):
        """Set locks to be used with a LockDistributor. Keys should be generated using get_lock_keys().

        Parameters
        ----------
        distributor : LockDistributor
            Lock distributor
        """
        self.set_locks_with_dict(distributor.get_locks(self.get_lock_keys()))

    def set_locks_with_dict(self, locks: Dict[str, Lock]):
        """Set locks to be used. Keys should be generated using get_lock_keys().

        Parameters
        ----------
        locks : Dict[str, Lock]
            Dict of locks with keys from get_lock_keys()
        """
        self._locks = locks

    @staticmethod
    def _get_lock_key(logical_id: str, api_call: ApiCallTypes) -> str:
        """Get a single lock key for a pair of resource and API call.

        Parameters
        ----------
        logical_id : str
            Logical ID of a resource.
        api_call : str
            API call the resource will use.

        Returns
        -------
        str
            String key created with logical ID and API call name.
        """
        return f"{logical_id}_{api_call.value}"

    def _get_lock_chain(self) -> LockChain:
        """Return a LockChain object for all the locks

        Returns
        -------
        Optional[LockChain]
            A LockChain object containing all locks. None if there are no locks.
        """
        if self._locks:
            return LockChain(self._locks)
        raise MissingLockException("Missing Locks for LockChain")

    def _get_resource(self, resource_identifier: str) -> Optional[Dict[str, Any]]:
        """Get a resource dict with resource identifier

        Parameters
        ----------
        resource_identifier : str
            Resource identifier

        Returns
        -------
        Optional[Dict[str, Any]]
            Resource dict containing its template fields.
        """
        return get_resource_by_id(self._stacks, ResourceIdentifier(resource_identifier)) if self._stacks else None

    def get_physical_id(self, resource_identifier: str) -> str:
        """Get the physical ID of a resource using physical_id_mapping. This does not directly check with remote.

        Parameters
        ----------
        resource_identifier : str
            Resource identifier

        Returns
        -------
        str
            Resource physical ID

        Raises
        ------
        MissingPhysicalResourceError
            Resource does not exist in the physical ID mapping.
            This could mean remote and local templates are not in sync.
        """
        physical_id = self._physical_id_mapping.get(resource_identifier)
        if not physical_id:
            raise MissingPhysicalResourceError(resource_identifier)

        return physical_id

    @abstractmethod
    def _equality_keys(self) -> Any:
        """This method needs to be overridden to distinguish between multiple instances of SyncFlows
        If the return values of two instances are the same, then those two instances will be assumed to be equal.

        Returns
        -------
        Any
            Anything that can be hashed and compared with "=="
        """
        raise NotImplementedError("_equality_keys is not implemented.")

    def __hash__(self) -> int:
        return hash((type(self), self._equality_keys()))

    def __eq__(self, o: object) -> bool:
        if type(o) is not type(self):
            return False
        return cast(bool, self._equality_keys() == cast(SyncFlow, o)._equality_keys())

    @property
    def log_name(self) -> str:
        """
        Returns
        -------
        str
            Human readable name/identifier for logging purposes
        """
        return self._log_name

    @property
    def log_prefix(self) -> str:
        """
        Returns
        -------
        str
            Log prefix to be used for logging.
        """
        return f"SyncFlow [{self.log_name}]: "

    def execute(self) -> List["SyncFlow"]:
        """Execute the sync flow and returns a list of dependent sync flows.
        Skips sync() and gather_dependencies() if compare() is True

        Returns
        -------
        List[SyncFlow]
            A list of dependent sync flows
        """
        dependencies: List["SyncFlow"] = list()
        LOG.debug("%sSetting Up", self.log_prefix)
        self.set_up()
        LOG.debug("%sGathering Resources", self.log_prefix)
        self.gather_resources()
        LOG.debug("%sComparing with Remote", self.log_prefix)
        if (not self.compare_local()) and (not self.compare_remote()):
            LOG.debug("%sSyncing", self.log_prefix)
            self.sync()
            LOG.debug("%sUpdating local hash of the sync flow", self.log_prefix)
            self._update_local_hash()
            LOG.debug("%sGathering Dependencies", self.log_prefix)
            dependencies = self.gather_dependencies()
        LOG.debug("%sFinished", self.log_prefix)
        return dependencies


def get_definition_path(
    resource: Dict, identifier: str, use_base_dir: bool, base_dir: str, stacks: List[Stack]
) -> Optional[Path]:
    """
    A helper method used by non-function sync flows to resolve definition file path
    that are relative to the child stack to absolute path for nested stacks

    Parameters
    -------
    resource: Dict
        The resource's template dict
    identifier: str
        The logical ID identifier of the resource
    use_base_dir: bool
        Whether or not the base_dir option was used
    base_dir: str
        Base directory if provided, otherwise the root template directory
    stacks: List[Stack]
        The list of stacks for the application

    Returns
    -------
    Optional[Path]
        A resolved absolute path for the definition file
    """
    properties = resource.get("Properties", {})
    definition_file = properties.get("DefinitionUri")
    definition_path = None
    if definition_file:
        definition_path = Path(base_dir).joinpath(definition_file)
        if not use_base_dir:
            child_stack = Stack.get_stack_by_full_path(ResourceIdentifier(identifier).stack_path, stacks)
            if child_stack:
                definition_path = Path(child_stack.location).parent.joinpath(definition_file)
    return definition_path



"""
Base classes that implement the CLI framework
"""
import importlib
import logging
from collections import OrderedDict

import click

from samcli.cli.formatters import RootCommandHelpTextFormatter
from samcli.cli.root.command_list import SAM_CLI_COMMANDS
from samcli.cli.row_modifiers import HighlightNewRowNameModifier, RowDefinition, ShowcaseRowModifier

logger = logging.getLogger(__name__)

# Commands that are bundled with the CLI by default in app life-cycle order.

_SAM_CLI_COMMAND_PACKAGES = [
    "samcli.commands.init",
    "samcli.commands.validate.validate",
    "samcli.commands.build",
    "samcli.commands.local.local",
    "samcli.commands.package",
    "samcli.commands.deploy",
    "samcli.commands.delete",
    "samcli.commands.logs",
    "samcli.commands.publish",
    "samcli.commands.traces",
    "samcli.commands.sync",
    "samcli.commands.pipeline.pipeline",
    "samcli.commands.list.list",
    # We intentionally do not expose the `bootstrap` command for now. We might open it up later
    # "samcli.commands.bootstrap",
]


class BaseCommand(click.MultiCommand):
    """
    Dynamically loads commands. It takes a list of names of Python packages representing the commands, loads
    these packages, and initializes them as Click commands. If a command "hello" is available in a Python package
    "foo.bar.hello", then this package name is passed to this class to load the command. This allows commands
    to be written as standalone packages that are dynamically initialized by the CLI.

    Each command, along with any subcommands, is implemented using Click annotations. When the command is loaded
    dynamically, this class expects the Click object to be exposed through an attribute called ``cli``. If the
    attribute is not present, or is not a Click object, then an exception will be raised.

    For example: if "foo.bar.hello" is the package where "hello" command is implemented, then
    "/foo/bar/hello/__init__.py" file is expected to contain a Click object called ``cli``.

    The command package is dynamically loaded using Python's standard ``importlib`` library. Therefore package names
    can be specified using the standard Python's dot notation such as "foo.bar.hello".

    By convention, the name of last module in the package's name is the command's name. ie. A package of "foo.bar.baz"
    will produce a command name "baz".
    """

    class CustomFormatterContext(click.Context):
        formatter_class = RootCommandHelpTextFormatter

    context_class = CustomFormatterContext

    def __init__(self, *args, cmd_packages=None, **kwargs):
        """
        Initializes the class, optionally with a list of available commands

        :param cmd_packages: List of Python packages names of CLI commands
        :param args: Other Arguments passed to super class
        :param kwargs: Other Arguments passed to super class
        """
        # alias -h to --help for all commands
        kwargs["context_settings"] = dict(help_option_names=["-h", "--help"])
        super().__init__(*args, **kwargs)

        if not cmd_packages:
            cmd_packages = _SAM_CLI_COMMAND_PACKAGES

        self._commands = {}
        self._commands = BaseCommand._set_commands(cmd_packages)

    @staticmethod
    def _set_commands(package_names):
        """
        Extract the command name from package name. Last part of the module path is the command
        ie. if path is foo.bar.baz, then "baz" is the command name.

        :param package_names: List of package names
        :return: Dictionary with command name as key and the package name as value.
        """

        commands = OrderedDict()

        for pkg_name in package_names:
            cmd_name = pkg_name.split(".")[-1]
            commands[cmd_name] = pkg_name

        return commands

    def format_options(self, ctx: click.Context, formatter: RootCommandHelpTextFormatter):  # type: ignore
        # NOTE(sriram-mv): `ignore` is put in place here for mypy even though it is the correct behavior,
        # as the `formatter_class` can be set in subclass of Command. If ignore is not set,
        # mypy raises argument needs to be HelpFormatter as super class defines it.
        # NOTE(sriram-mv): Re-order options so that they come after the commands.
        self.format_commands(ctx, formatter)
        opts = [RowDefinition(name="", text="\n")]
        for param in self.get_params(ctx):
            row = param.get_help_record(ctx)
            if row is not None:
                term, help_text = row
                opts.append(RowDefinition(name=term, text=help_text))

        if opts:
            with formatter.indented_section(name="Options", extra_indents=1):
                formatter.write_rd(opts)

        with formatter.indented_section(name="Examples", extra_indents=1):
            formatter.write_rd(
                [
                    RowDefinition(
                        name="",
                        text="\n",
                    ),
                    RowDefinition(
                        name="Get Started:",
                        text=click.style(f"${ctx.command_path} init"),
                        extra_row_modifiers=[ShowcaseRowModifier()],
                    ),
                ],
            )

    def format_commands(self, ctx: click.Context, formatter: RootCommandHelpTextFormatter):  # type: ignore
        # NOTE(sriram-mv): `ignore` is put in place here for mypy even though it is the correct behavior,
        # as the `formatter_class` can be set in subclass of Command. If ignore is not set,
        # mypy raises argument needs to be HelpFormatter as super class defines it.
        with formatter.section("Commands"):
            with formatter.section("Create an App"):
                formatter.write_rd(
                    [
                        RowDefinition(name="init", text=SAM_CLI_COMMANDS.get("init", "")),
                    ],
                )

            with formatter.section("Develop your App"):
                formatter.write_rd(
                    [
                        RowDefinition(
                            name="build",
                            text=SAM_CLI_COMMANDS.get("build", ""),
                        ),
                        RowDefinition(
                            name="local",
                            text=SAM_CLI_COMMANDS.get("local", ""),
                        ),
                        RowDefinition(
                            name="validate",
                            text=SAM_CLI_COMMANDS.get("validate", ""),
                        ),
                        RowDefinition(
                            name="sync",
                            text=SAM_CLI_COMMANDS.get("sync", ""),
                            extra_row_modifiers=[HighlightNewRowNameModifier()],
                        ),
                    ],
                )

            with formatter.section("Deploy your App"):
                formatter.write_rd(
                    [
                        RowDefinition(
                            name="package",
                            text=SAM_CLI_COMMANDS.get("package", ""),
                        ),
                        RowDefinition(
                            name="deploy",
                            text=SAM_CLI_COMMANDS.get("deploy", ""),
                        ),
                    ]
                )

            with formatter.section("Monitor your App"):
                formatter.write_rd(
                    [
                        RowDefinition(
                            name="logs",
                            text=SAM_CLI_COMMANDS.get("logs", ""),
                        ),
                        RowDefinition(
                            name="traces",
                            text=SAM_CLI_COMMANDS.get("traces", ""),
                        ),
                    ],
                )

            with formatter.section("And More"):
                formatter.write_rd(
                    [
                        RowDefinition(
                            name="list",
                            text=SAM_CLI_COMMANDS.get("list", ""),
                            extra_row_modifiers=[HighlightNewRowNameModifier()],
                        ),
                        RowDefinition(
                            name="delete",
                            text=SAM_CLI_COMMANDS.get("delete", ""),
                        ),
                        RowDefinition(
                            name="pipeline",
                            text=SAM_CLI_COMMANDS.get("pipeline", ""),
                        ),
                        RowDefinition(
                            name="publish",
                            text=SAM_CLI_COMMANDS.get("publish", ""),
                        ),
                    ],
                )

    def list_commands(self, ctx):
        """
        Overrides a method from Click that returns a list of commands available in the CLI.

        :param ctx: Click context
        :return: List of commands available in the CLI
        """
        return list(self._commands.keys())

    def get_command(self, ctx, cmd_name):
        """
        Overrides method from ``click.MultiCommand`` that returns Click CLI object for given command name, if found.

        :param ctx: Click context
        :param cmd_name: Top-level command name
        :return: Click object representing the command
        """
        if cmd_name not in self._commands:
            logger.error("Command %s not available", cmd_name)
            return None

        pkg_name = self._commands[cmd_name]

        mod = None
        try:
            if ctx.obj:
                # NOTE(sriram-mv): Only attempt to import if a relevant `aws sam cli` context has been set.
                # `aws sam cli` context is only set after the `samcli.cli.main:cli` has been executed.
                mod = importlib.import_module(pkg_name)
        except ImportError:
            logger.exception("Command '%s' is not configured correctly. Unable to import '%s'", cmd_name, pkg_name)
            return None

        if mod is not None:
            if not hasattr(mod, "cli"):
                logger.error(
                    "Command %s is not configured correctly. It must expose an function called 'cli'", cmd_name
                )
                return None

        return mod.cli if mod else click.Command(name=cmd_name, short_help=SAM_CLI_COMMANDS.get(cmd_name, ""))



"""
Custom Click options for multiple arguments
"""

import click


class OptionNargs(click.Option):
    """
    A custom option class that allows parsing for multiple arguments
    for an option, when the number of arguments for an option are unknown.
    """

    def __init__(self, *args, **kwargs):
        self.nargs = kwargs.pop("nargs", -1)
        super().__init__(*args, **kwargs)
        self._previous_parser_process = None
        self._nargs_parser = None

    def add_to_parser(self, parser, ctx):
        def parser_process(value, state):
            # look ahead into arguments till we reach the next option.
            # the next option starts with a prefix which is either '-' or '--'
            next_option = False
            value = [value]

            while state.rargs and not next_option:
                for prefix in self._nargs_parser.prefixes:
                    if state.rargs[0].startswith(prefix):
                        next_option = True
                if not next_option:
                    value.append(state.rargs.pop(0))

            value = tuple(value)

            # call the actual process
            self._previous_parser_process(value, state)

        # Add current option to Parser by calling add_to_parser on the super class.
        super().add_to_parser(parser, ctx)
        for name in self.opts:
            # Get OptionParser object for current option
            option_parser = getattr(parser, "_long_opt").get(name) or getattr(parser, "_short_opt").get(name)
            if option_parser:
                # Monkey patch `process` method for click.parser.Option class.
                # This allows for setting multiple parsed values into current option arguments
                self._nargs_parser = option_parser
                self._previous_parser_process = option_parser.process
                option_parser.process = parser_process
                break



"""
Contains information about newer version checker for SAM CLI
"""
import logging
from datetime import datetime, timedelta
from functools import wraps

import click
from requests import get

from samcli import __version__ as installed_version
from samcli.cli.global_config import GlobalConfig

LOG = logging.getLogger(__name__)

AWS_SAM_CLI_PYPI_ENDPOINT = "https://pypi.org/pypi/aws-sam-cli/json"
AWS_SAM_CLI_INSTALL_DOCS = (
    "https://docs.aws.amazon.com/serverless-application-model/latest/developerguide/serverless-sam-cli-install.html"
)
PYPI_CALL_TIMEOUT_IN_SECONDS = 5
DELTA_DAYS = 7


def check_newer_version(func):
    """
    This function returns a wrapped function definition, which checks if there are newer version of SAM CLI available

    Parameters
    ----------
    func: function reference
        Actual function (command) which will be executed

    Returns
    -------
    function reference:
        A wrapped function reference which executes original function and checks newer version of SAM CLI
    """

    @wraps(func)
    def wrapped(*args, **kwargs):
        # execute actual command first
        actual_result = func(*args, **kwargs)
        # check and inform newer version if it is available
        _inform_newer_version()

        return actual_result

    return wrapped


def _inform_newer_version(force_check=False) -> None:
    """
    Compares installed SAM CLI version with the up to date version from PyPi,
    and print information if up to date version is different then what is installed now

    It will store last version check time into GlobalConfig, so that it won't be running all the time
    Currently, it will be checking weekly

    Parameters
    ----------
    force_check: bool
        When it is True, it will trigger checking new version of SAM CLI. Default value is False

    """
    # run everything else in try-except block
    global_config = None
    need_to_update_last_check_time = True
    try:
        global_config = GlobalConfig()
        last_version_check = global_config.last_version_check

        if force_check or is_version_check_overdue(last_version_check):
            fetch_and_compare_versions()
        else:
            need_to_update_last_check_time = False
    except Exception as e:
        LOG.debug("New version check failed", exc_info=e)
    finally:
        if need_to_update_last_check_time:
            update_last_check_time()


def fetch_and_compare_versions() -> None:
    """
    Compare current up to date version with the installed one, and inform if a newer version available
    """
    response = get(AWS_SAM_CLI_PYPI_ENDPOINT, timeout=PYPI_CALL_TIMEOUT_IN_SECONDS)
    result = response.json()
    latest_version = result.get("info", {}).get("version", None)
    LOG.debug("Installed version %s, current version %s", installed_version, latest_version)
    if latest_version and installed_version != latest_version:
        click.secho(
            f"\nSAM CLI update available ({latest_version}); ({installed_version} installed)", fg="green", err=True
        )
        click.echo(f"To download: {AWS_SAM_CLI_INSTALL_DOCS}", err=True)


def update_last_check_time() -> None:
    """
    Update last_check_time in GlobalConfig
    """
    try:
        gc = GlobalConfig()
        gc.last_version_check = datetime.utcnow().timestamp()
    except Exception as e:
        LOG.debug("Updating last version check time was failed", exc_info=e)


def is_version_check_overdue(last_version_check) -> bool:
    """
    Check if last version check have been made longer then a week ago

    Parameters
    ----------
    last_version_check: epoch time
        last_version_check epoch time read from GlobalConfig

    Returns
    -------
    bool:
        True if last_version_check is None or older then a week, False otherwise
    """
    if last_version_check is None or type(last_version_check) not in [int, float]:
        return True

    epoch_week_ago = datetime.utcnow() - timedelta(days=DELTA_DAYS)
    return datetime.utcfromtimestamp(last_version_check) < epoch_week_ago



""" This module parses a json/yaml file that defines a flow of questions to fulfill the cookiecutter context"""
from typing import Dict, Optional, Tuple

import yaml

from samcli.commands.exceptions import UserException
from samcli.yamlhelper import parse_yaml_file

from .interactive_flow import InteractiveFlow
from .question import Question, QuestionFactory


class QuestionsNotFoundException(UserException):
    pass


class QuestionsFailedParsingException(UserException):
    pass


class InteractiveFlowCreator:
    @staticmethod
    def create_flow(flow_definition_path: str, extra_context: Optional[Dict] = None) -> InteractiveFlow:
        """
        This method parses the given json/yaml file to create an InteractiveFLow. It expects the file to define
        a list of questions. It parses the questions and add it to the flow in the same order they are defined
        in the file, i.e. the default next-question of a given question will be the next one defined in the file,
        while also respecting the question-defined _next_question map if provided.

        Parameters:
        ----------
        flow_definition_path: str
            A path to a json/yaml file that defines the questions of the flow. the file is expected to be in the
            following format:
            {
                "questions": [
                    {
                      "key": "key of the corresponding cookiecutter config",
                      "question": "the question to prompt to the user",
                      "kind": "the kind of the question, for example: confirm",
                      "isRequired": true/false,
                      # optional branching logic to jump to a particular question based on the answer. if not given
                      # will automatically go to next question
                      "nextQuestion": {
                        "True": "key of the question to jump to if the user answered 'Yes'",
                        "False": "key of the question to jump to if the user answered 'Yes'",
                      }
                      "default": "default_answer",
                      # the default value can also be loaded from cookiecutter context
                      # with a key path whose key path item can be loaded from cookiecutter as well.
                      "default": {
                        "keyPath": [
                            {
                                "valueOf": "key-of-another-question"
                            },
                            "pipeline_user"
                        ]
                      }
                      # assuming the answer of "key-of-another-question" is "ABC"
                      # the default value will be load from cookiecutter context with key "['ABC', 'pipeline_user]"
                    },
                    ...
                ]
            }
        extra_context: Dict
            if the template contains variable($variableName) this parameter provides the values for those variables.

        Returns: InteractiveFlow(questions={k1: q1, k2: q2, ...}, first_question_key="first question's key")
        """
        questions, first_question_key = InteractiveFlowCreator._load_questions(flow_definition_path, extra_context)
        return InteractiveFlow(questions=questions, first_question_key=first_question_key)

    @staticmethod
    def _load_questions(
        flow_definition_path: str, extra_context: Optional[Dict] = None
    ) -> Tuple[Dict[str, Question], str]:
        previous_question: Optional[Question] = None
        first_question_key: str = ""
        questions: Dict[str, Question] = {}
        questions_definition = InteractiveFlowCreator._parse_questions_definition(flow_definition_path, extra_context)

        try:
            for question in questions_definition.get("questions", []):
                q = QuestionFactory.create_question_from_json(question)
                if not first_question_key:
                    first_question_key = q.key
                elif previous_question and not previous_question.default_next_question_key:
                    previous_question.set_default_next_question_key(q.key)
                questions[q.key] = q
                previous_question = q
            return questions, first_question_key
        except (KeyError, ValueError, AttributeError, TypeError) as ex:
            raise QuestionsFailedParsingException(f"Failed to parse questions: {str(ex)}") from ex

    @staticmethod
    def _parse_questions_definition(file_path: str, extra_context: Optional[Dict] = None) -> Dict:
        """
        Read the questions definition file, do variable substitution, parse it as JSON/YAML

        Parameters
        ----------
        file_path : string
            Path to the questions definition to read
        extra_context : Dict
            if the file contains variable($variableName) this parameter provides the values for those variables.

        Raises
        ------
        QuestionsNotFoundException: if the file_path doesn't exist
        QuestionsFailedParsingException: if any error occurred during variables substitution or content parsing

        Returns
        -------
        questions data as a dictionary
        """

        try:
            return parse_yaml_file(file_path=file_path, extra_context=extra_context)
        except FileNotFoundError as ex:
            raise QuestionsNotFoundException(f"questions definition file not found at {file_path}") from ex
        except (KeyError, ValueError, yaml.YAMLError) as ex:
            raise QuestionsFailedParsingException(f"Failed to parse questions: {str(ex)}") from ex



"""
The symbol table that is used in IntrinsicResolver in order to resolve runtime attributes
"""
import logging
import os

from samcli.lib.intrinsic_resolver.intrinsic_property_resolver import IntrinsicResolver
from samcli.lib.intrinsic_resolver.invalid_intrinsic_exception import InvalidSymbolException

LOG = logging.getLogger(__name__)


class IntrinsicsSymbolTable:
    AWS_ACCOUNT_ID = "AWS::AccountId"
    AWS_NOTIFICATION_ARN = "AWS::NotificationArn"
    AWS_PARTITION = "AWS::Partition"
    AWS_REGION = "AWS::Region"
    AWS_STACK_ID = "AWS::StackId"
    AWS_STACK_NAME = "AWS::StackName"
    AWS_URL_PREFIX = "AWS::URLSuffix"
    AWS_NOVALUE = "AWS::NoValue"
    SUPPORTED_PSEUDO_TYPES = [
        AWS_ACCOUNT_ID,
        AWS_NOTIFICATION_ARN,
        AWS_PARTITION,
        AWS_REGION,
        AWS_STACK_ID,
        AWS_STACK_NAME,
        AWS_URL_PREFIX,
        AWS_NOVALUE,
    ]

    # There is not much benefit in infering real values for these parameters in local development context. These values
    # are usually representative of an AWS environment and stack, but in local development scenario they don't make
    # sense. If customers choose to, they can always override this value through the CLI interface.
    DEFAULT_PSEUDO_PARAM_VALUES = {
        "AWS::AccountId": "123456789012",
        "AWS::Partition": "aws",
        "AWS::Region": "us-east-1",
        "AWS::StackName": "local",
        "AWS::StackId": "arn:aws:cloudformation:us-east-1:123456789012:stack/"
        "local/51af3dc0-da77-11e4-872e-1234567db123",
        "AWS::URLSuffix": "localhost",
    }

    REGIONS = {
        "us-east-1": ["us-east-1a", "us-east-1b", "us-east-1c", "us-east-1d", "us-east-1e", "us-east-1f"],
        "us-west-1": ["us-west-1b", "us-west-1c"],
        "eu-north-1": ["eu-north-1a", "eu-north-1b", "eu-north-1c"],
        "ap-northeast-3": ["ap-northeast-3a"],
        "ap-northeast-2": ["ap-northeast-2a", "ap-northeast-2b", "ap-northeast-2c"],
        "ap-northeast-1": ["ap-northeast-1a", "ap-northeast-1c", "ap-northeast-1d"],
        "sa-east-1": ["sa-east-1a", "sa-east-1c"],
        "ap-southeast-1": ["ap-southeast-1a", "ap-southeast-1b", "ap-southeast-1c"],
        "ca-central-1": ["ca-central-1a", "ca-central-1b"],
        "ap-southeast-2": ["ap-southeast-2a", "ap-southeast-2b", "ap-southeast-2c"],
        "us-west-2": ["us-west-2a", "us-west-2b", "us-west-2c", "us-west-2d"],
        "us-east-2": ["us-east-2a", "us-east-2b", "us-east-2c"],
        "ap-south-1": ["ap-south-1a", "ap-south-1b", "ap-south-1c"],
        "eu-central-1": ["eu-central-1a", "eu-central-1b", "eu-central-1c"],
        "eu-west-1": ["eu-west-1a", "eu-west-1b", "eu-west-1c"],
        "eu-west-2": ["eu-west-2a", "eu-west-2b", "eu-west-2c"],
        "eu-west-3": ["eu-west-3a", "eu-west-3b", "eu-west-3c"],
        "cn-north-1": [],
        "us-gov-west-1": [],
    }

    DEFAULT_PARTITION = "aws"
    GOV_PARTITION = "aws-us-gov"
    CHINA_PARTITION = "aws-cn"
    CHINA_PREFIX = "cn"
    GOV_PREFIX = "gov"

    CHINA_URL_PREFIX = "amazonaws.com.cn"
    DEFAULT_URL_PREFIX = "amazonaws.com"

    AWS_NOTIFICATION_SERVICE_NAME = "sns"
    ARN_SUFFIX = ".Arn"

    CFN_RESOURCE_TYPE = "Type"
    CFN_RESOURCE_PROPERTIES = "Properties"
    CFN_LAMBDA_FUNCTION_NAME = "FunctionName"

    def __init__(
        self, template=None, logical_id_translator=None, default_type_resolver=None, common_attribute_resolver=None
    ):
        """
        Initializes the Intrinsic Symbol Table so that runtime attributes can be resolved.

        The code is defaulted in the following order logical_id_translator => parameters => default_type_resolver =>
        common_attribute_resolver

        If the item is a pseudo type, it will run through the logical_id_translator and if it doesn't exist there
        it will generate a default one and save it in the logical_id_translator as a cache for future computation.
        Parameters
        ------------
        template : Optional[Dict]
            An optional dictionary representing the template
        logical_id_translator : dict
            This will act as the default symbol table resolver. The resolver will first check if the attribute is
            explicitly defined in this dictionary and do the relevant translation.

            All Logical Ids and Pseudo types can be included here.
            {
                "RestApi.Test": {  # this could be used with RestApi.Deployment => NewRestApi
                    "Ref": "NewRestApi"
                },
                "LambdaFunction": {
                    "Ref": "LambdaFunction",
                    "Arn": "MyArn"
                }
                "AWS::Region": "us-east-1"
            }
        default_type_resolver : dict
            This can be used provide common attributes that are true across all objects of a certain type.
            This can be in the format of
            {
                "AWS::ApiGateway::RestApi": {
                    "RootResourceId": "/"
                }
            }
            or can also be a function that takes in (logical_id, attribute_type) => string
            {
                "AWS::ApiGateway::RestApi": {
                    "RootResourceId": (lambda l, a, p, r: p.get("ResourceId"))
                }
            }
        common_attribute_resolver : dict
            This is a clean way of specifying common attributes across all types.
            The value can either be a function of the form string or (logical_id) => string
            {
                "Ref": lambda p,r: "",
                "Arn:": arn_resolver
            }
        """
        self.logical_id_translator = logical_id_translator or {}

        self._template = template or {}
        self._parameters = self._template.get("Parameters", {})
        self._resources = self._template.get("Resources", {})

        self.default_type_resolver = default_type_resolver or self.get_default_type_resolver()
        self.common_attribute_resolver = common_attribute_resolver or self.get_default_attribute_resolver()
        self.default_pseudo_resolver = self.get_default_pseudo_resolver()

    def get_default_pseudo_resolver(self):
        return {
            IntrinsicsSymbolTable.AWS_ACCOUNT_ID: self.handle_pseudo_account_id,
            IntrinsicsSymbolTable.AWS_PARTITION: self.handle_pseudo_partition,
            IntrinsicsSymbolTable.AWS_REGION: self.handle_pseudo_region,
            IntrinsicsSymbolTable.AWS_STACK_ID: self.handle_pseudo_stack_id,
            IntrinsicsSymbolTable.AWS_STACK_NAME: self.handle_pseudo_stack_name,
            IntrinsicsSymbolTable.AWS_NOVALUE: self.handle_pseudo_no_value,
            IntrinsicsSymbolTable.AWS_URL_PREFIX: self.handle_pseudo_url_prefix,
        }

    def get_default_attribute_resolver(self):
        return {"Ref": lambda logical_id: logical_id, "Arn": self.arn_resolver}

    @staticmethod
    def get_default_type_resolver():
        return {
            "AWS::ApiGateway::RestApi": {
                "RootResourceId": "/"  # It usually used as a reference to the parent id of the RestApi,
            },
            "AWS::Lambda::LayerVersion": {
                IntrinsicResolver.REF: lambda logical_id: {IntrinsicResolver.REF: logical_id}
            },
            "AWS::Serverless::LayerVersion": {
                IntrinsicResolver.REF: lambda logical_id: {IntrinsicResolver.REF: logical_id}
            },
        }

    def resolve_symbols(self, logical_id, resource_attribute, ignore_errors=False):
        """
        This function resolves all the symbols given a logical id and a resource_attribute for Fn::GetAtt and Ref.
        This boils Ref into a type of Fn:GetAtt to simplify the implementation.
        For example:
            {"Ref": "AWS::REGION"} => resolve_symbols("AWS::REGION", "REF")
            {"Fn::GetAtt": ["logical_id", "attribute_type"] => resolve_symbols(logical_id, attribute_type)


        First pseudo types are checked. If item is present in the logical_id_translator it is returned.
        Otherwise, it falls back to the default_pseudo_resolver

        Then the default_type_resolver is checked, which has common attributes and functions for each types.
        Then the common_attribute_resolver is run, which has functions that are common for each attribute.
        Parameters
        -----------
        logical_id: str
            The logical id of the resource in question or a pseudo type.
        resource_attribute: str
            The resource attribute of the resource in question or Ref for psuedo types.
        ignore_errors: bool
            An optional flags to not return errors. This used in sub

        Return
        -------
        This resolves the attribute
        """
        # pylint: disable-msg=too-many-return-statements
        translated = self.get_translation(logical_id, resource_attribute)
        if translated:
            return translated

        if logical_id in self.SUPPORTED_PSEUDO_TYPES:
            translated = self.default_pseudo_resolver.get(logical_id)()
            self.logical_id_translator[logical_id] = translated
            return translated

        # Handle Default Parameters
        translated = self._parameters.get(logical_id, {}).get("Default")
        if translated is not None:
            return translated
        # Handle Default Property Type Resolution
        resource_type = self._resources.get(logical_id, {}).get(IntrinsicsSymbolTable.CFN_RESOURCE_TYPE)

        resolver = self.default_type_resolver.get(resource_type, {}).get(resource_attribute) if resource_type else {}
        if resolver:
            if callable(resolver):
                return resolver(logical_id)
            return resolver

        # Handle Attribute Type Resolution
        attribute_resolver = self.common_attribute_resolver.get(resource_attribute, {})
        if attribute_resolver:
            if callable(attribute_resolver):
                return attribute_resolver(logical_id)
            return attribute_resolver

        if ignore_errors:
            return "${}".format(logical_id + "." + resource_attribute)
        raise InvalidSymbolException(
            "The {} is not supported in the logical_id_translator, default_type_resolver, or the attribute_resolver."
            " It is also not a supported pseudo function".format(logical_id + "." + resource_attribute)
        )

    def arn_resolver(self, logical_id, service_name="lambda"):
        """
        This function resolves Arn in the format
            arn:{partition_name}:{service_name}:{aws_region}:{account_id}:{function_name}

        Parameters
        -----------
        logical_id: str
            This the reference to the function name used
        service_name: str
            This is the service name used such as lambda or sns

        Return
        -------
        The resolved Arn
        """
        aws_region = self.handle_pseudo_region()
        account_id = (
            self.logical_id_translator.get(IntrinsicsSymbolTable.AWS_ACCOUNT_ID) or self.handle_pseudo_account_id()
        )
        partition_name = self.handle_pseudo_partition()
        if service_name == "lambda":
            resource_name = self._get_function_name(logical_id)
            resource_name = self.logical_id_translator.get(resource_name) or resource_name

            str_format = "arn:{partition_name}:{service_name}:{aws_region}:{account_id}:function:{resource_name}"
        else:
            resource_name = logical_id
            resource_name = self.logical_id_translator.get(resource_name) or resource_name

            str_format = "arn:{partition_name}:{service_name}:{aws_region}:{account_id}:{resource_name}"

        return str_format.format(
            partition_name=partition_name,
            service_name=service_name,
            aws_region=aws_region,
            account_id=account_id,
            resource_name=resource_name,
        )

    def _get_function_name(self, logical_id):
        """
        This function returns the function name associated with the logical ID.
        If the template doesn't define a FunctionName, it will just return the
        logical ID, which is the default function name.

        Parameters
        -----------
        logical_id: str
            This the reference to the function name used

        Return
        -------
        The function name
        """
        if not self._resources:
            return logical_id
        resource_definition_dict = self._resources.get(logical_id)
        if not resource_definition_dict:
            return logical_id

        resource_properties = resource_definition_dict.get(IntrinsicsSymbolTable.CFN_RESOURCE_PROPERTIES)
        if not resource_properties:
            return logical_id

        resource_name = resource_properties.get(IntrinsicsSymbolTable.CFN_LAMBDA_FUNCTION_NAME)
        return resource_name or logical_id

    def get_translation(self, logical_id, resource_attributes=IntrinsicResolver.REF):
        """
        This gets the logical_id_translation of the logical id and resource_attributes.

        Parameters
        ----------
        logical_id: str
            This is the logical id of the resource in question
        resource_attributes: str
            This is the attribute required. By default, it is a REF type

        Returns
        --------
        This returns the translated item if it already exists

        """
        logical_id_item = self.logical_id_translator.get(logical_id, {})
        if any(isinstance(logical_id_item, object_type) for object_type in [str, list, bool, int]):
            if resource_attributes not in (IntrinsicResolver.REF, ""):
                return None
            return logical_id_item

        return logical_id_item.get(resource_attributes)

    @staticmethod
    def get_availability_zone(region):
        """
        This gets the availability zone from the the specified region

        Parameters
        -----------
        region: str
            The specified region from the SymbolTable region

        Return
        -------
        The list of availability zones for the specified region
        """
        return IntrinsicsSymbolTable.REGIONS.get(region)

    @staticmethod
    def handle_pseudo_account_id():
        """
        This gets a default account id from SamBaseProvider.
        Return
        -------
        A pseudo account id
        """
        return IntrinsicsSymbolTable.DEFAULT_PSEUDO_PARAM_VALUES.get(IntrinsicsSymbolTable.AWS_ACCOUNT_ID)

    def handle_pseudo_region(self):
        """
        Gets the region from the environment and defaults to a the default region from the global variables.

        This is only run if it is not specified by the logical_id_translator as a default.

        Return
        -------
        The region from the environment or a default one
        """
        return (
            self.logical_id_translator.get(IntrinsicsSymbolTable.AWS_REGION)
            or os.getenv("AWS_REGION")
            or IntrinsicsSymbolTable.DEFAULT_PSEUDO_PARAM_VALUES.get(IntrinsicsSymbolTable.AWS_REGION)
        )

    def handle_pseudo_url_prefix(self):
        """
        This gets the AWS::UrlSuffix for the intrinsic with the china and regular prefix.

        This is only run if it is not specified by the logical_id_translator as a default.
        Return
        -------
        The url prefix of amazonaws.com or amazonaws.com.cn
        """
        aws_region = self.handle_pseudo_region()
        if self.CHINA_PREFIX in aws_region:
            return self.CHINA_URL_PREFIX
        return self.DEFAULT_URL_PREFIX

    def handle_pseudo_partition(self):
        """
        This resolves AWS::Partition so that the correct partition is returned depending on the region.

        This is only run if it is not specified by the logical_id_translator as a default.

        Return
        -------
        A pseudo partition like aws-cn or aws or aws-gov
        """
        aws_region = self.handle_pseudo_region()
        if self.CHINA_PREFIX in aws_region:
            return self.CHINA_PARTITION
        if self.GOV_PREFIX in aws_region:
            return self.GOV_PARTITION
        return self.DEFAULT_PARTITION

    @staticmethod
    def handle_pseudo_stack_id():
        """
        This resolves AWS::StackId by using the SamBaseProvider as the default value.

        This is only run if it is not specified by the logical_id_translator as a default.

        Return
        -------
        A randomized string
        """
        return IntrinsicsSymbolTable.DEFAULT_PSEUDO_PARAM_VALUES.get(IntrinsicsSymbolTable.AWS_STACK_ID)

    @staticmethod
    def handle_pseudo_stack_name():
        """
        This resolves AWS::StackName by using the SamBaseProvider as the default value.

        This is only run if it is not specified by the logical_id_translator as a default.

        Return
        -------
        A randomized string
        """
        return IntrinsicsSymbolTable.DEFAULT_PSEUDO_PARAM_VALUES.get(IntrinsicsSymbolTable.AWS_STACK_NAME)

    @staticmethod
    def handle_pseudo_no_value():
        """
        This resolves AWS::NoValue so that it returns the python None
        """
        return None



"""
Docker container related exceptions
"""


class ContainerNotStartableException(Exception):
    pass


class NoFreePortsError(Exception):
    """
    Exception to raise when there are no free ports found in a specified range.
    """



"""
Container for a mapper and a consumer
"""
from dataclasses import dataclass

from samcli.lib.list.list_interfaces import ListInfoPullerConsumer, Mapper


@dataclass
class MapperConsumerContainer:
    mapper: Mapper
    consumer: ListInfoPullerConsumer



"""
Exceptions that are raised by sam package
"""
from samcli.commands.exceptions import UserException


class InvalidLocalPathError(UserException):
    def __init__(self, resource_id, property_name, local_path):
        self.resource_id = resource_id
        self.property_name = property_name
        self.local_path = local_path
        message_fmt = (
            "Parameter {property_name} of resource {resource_id} refers "
            "to a file or folder that does not exist {local_path}"
        )
        super().__init__(
            message=message_fmt.format(
                resource_id=self.resource_id, property_name=self.property_name, local_path=self.local_path
            )
        )


class InvalidTemplateUrlParameterError(UserException):
    def __init__(self, resource_id, property_name, template_path):
        self.resource_id = resource_id
        self.property_name = property_name
        self.template_path = template_path

        message_fmt = (
            "{property_name} parameter of {resource_id} resource is invalid. "
            "It must be a S3 URL or path to CloudFormation "
            "template file. Actual: {template_path}"
        )
        super().__init__(
            message=message_fmt.format(
                property_name=self.property_name, resource_id=self.resource_id, template_path=self.template_path
            )
        )


class ExportFailedError(UserException):
    def __init__(self, resource_id, property_name, property_value, ex):
        self.resource_id = resource_id
        self.property_name = property_name
        self.property_value = property_value
        self.ex = ex

        message_fmt = (
            "Unable to upload artifact {property_value} referenced "
            "by {property_name} parameter of {resource_id} resource."
            "\n"
            "{ex}"
        )

        super().__init__(
            message=message_fmt.format(
                property_value=self.property_value,
                property_name=self.property_name,
                resource_id=self.resource_id,
                ex=self.ex,
            )
        )


class DeleteArtifactFailedError(UserException):
    def __init__(self, resource_id, property_name, ex):
        self.resource_id = resource_id
        self.property_name = property_name
        self.ex = ex

        message_fmt = (
            "Unable to delete artifact referenced "
            "by {property_name} parameter of {resource_id} resource."
            "\n"
            "{ex}"
        )

        super().__init__(
            message=message_fmt.format(
                property_name=self.property_name,
                resource_id=self.resource_id,
                ex=self.ex,
            )
        )


class ImageNotFoundError(UserException):
    def __init__(self, resource_id, property_name, message_fmt):
        self.resource_id = resource_id
        self.property_name = property_name

        super().__init__(
            message=message_fmt.format(
                property_name=self.property_name,
                resource_id=self.resource_id,
            )
        )


class ECRAuthorizationError(UserException):
    def __init__(self, msg):
        self.msg = msg
        super().__init__(message=self.msg)


class DockerLoginFailedError(UserException):
    def __init__(self, msg):
        self.msg = msg
        super().__init__(message=self.msg)


class DockerPushFailedError(UserException):
    def __init__(self, msg):
        self.msg = msg
        super().__init__(message=self.msg)


class DockerGetLocalImageFailedError(UserException):
    def __init__(self, msg):
        self.msg = msg
        super().__init__(message=self.msg)


class PackageFailedError(UserException):
    def __init__(self, template_file, ex):
        self.template_file = template_file
        self.ex = ex

        message_fmt = "Failed to package template: {template_file}. \n {ex}"

        super().__init__(message=message_fmt.format(template_file=self.template_file, ex=self.ex))


class NoSuchBucketError(UserException):
    def __init__(self, **kwargs):
        self.kwargs = kwargs

        message_fmt = "\nS3 Bucket does not exist."

        super().__init__(message=message_fmt.format(**self.kwargs))


class BucketNotSpecifiedError(UserException):
    def __init__(self, **kwargs):
        self.kwargs = kwargs

        message_fmt = "\nS3 Bucket not specified, use --s3-bucket to specify a bucket name, or use --resolve-s3 \
to create a managed default bucket, or run sam deploy --guided"

        super().__init__(message=message_fmt.format(**self.kwargs))


class PackageResolveS3AndS3SetError(UserException):
    def __init__(self):
        message_fmt = "Cannot use both --resolve-s3 and --s3-bucket parameters. Please use only one."

        super().__init__(message=message_fmt)


class PackageResolveS3AndS3NotSetError(UserException):
    def __init__(self):
        message_fmt = "Cannot skip both --resolve-s3 and --s3-bucket parameters. Please provide one of these arguments."

        super().__init__(message=message_fmt)



""" Application Environment """
import hashlib
import json
import os
import pathlib
import re
import socket
from itertools import chain
from typing import Dict, List, Optional, Tuple
from urllib.parse import urlparse

import boto3
import click
import requests
from botocore.exceptions import ClientError
from OpenSSL import SSL, crypto  # type: ignore

from samcli.commands.pipeline.bootstrap.guided_context import BITBUCKET, GITHUB_ACTIONS, GITLAB, OPEN_ID_CONNECT
from samcli.commands.pipeline.bootstrap.pipeline_oidc_provider import PipelineOidcProvider
from samcli.lib.config.samconfig import SamConfig
from samcli.lib.pipeline.bootstrap.resource import ECRImageRepository, IAMUser, OidcProvider, Resource
from samcli.lib.utils.colors import Colored
from samcli.lib.utils.managed_cloudformation_stack import StackOutput, update_stack

CFN_TEMPLATE_PATH = str(pathlib.Path(os.path.dirname(__file__)))
STACK_NAME_PREFIX = "aws-sam-cli-managed"
STAGE_RESOURCES_STACK_NAME_SUFFIX = "pipeline-resources"
STAGE_RESOURCES_CFN_TEMPLATE = "stage_resources.yaml"
PIPELINE_USER = "pipeline_user"
PIPELINE_EXECUTION_ROLE = "pipeline_execution_role"
CLOUDFORMATION_EXECUTION_ROLE = "cloudformation_execution_role"
ARTIFACTS_BUCKET = "artifacts_bucket"
ECR_IMAGE_REPOSITORY = "image_repository"
OIDC_PROVIDER_URL = "oidc_provider_url"
OIDC_CLIENT_ID = "oidc_client_id"
OIDC_PROVIDER = "oidc_provider"
GITHUB_ORG = "github_org"
GITHUB_REPO = "github_repo"
GITLAB_GROUP = "gitlab_group"
GITLAB_PROJECT = "gitlab_project"
DEPLOYMENT_BRANCH = "deployment_branch"
BITBUCKET_REPO_UUID = "bitbucket_repo_uuid"
PERMISSIONS_PROVIDER = "permissions_provider"
OIDC_SUPPORTED_PROVIDER = [GITHUB_ACTIONS, GITLAB, BITBUCKET]
REGION = "region"


def _get_secure_ssl_context() -> SSL.Context:
    """
    Returns a SSL Context with secure settings
    """
    ctx = SSL.Context(SSL.TLS_METHOD)
    ctx.set_options(SSL.OP_NO_TLSv1)
    ctx.set_options(SSL.OP_NO_TLSv1_1)
    ctx.set_options(SSL.OP_NO_SSLv2)
    ctx.set_options(SSL.OP_NO_SSLv3)
    return ctx


class Stage:
    """
    Represents an application stage: Beta, Gamma, Prod ...etc

    Attributes
    ----------
    name: str
        The name of the environment
    aws_profile: Optional[str]
        The named AWS profile (in user's machine) of the AWS account to deploy this environment to.
    aws_region: Optional[str]
        The AWS region to deploy this environment to.
    pipeline_user: IAMUser
        The IAM User having its AccessKeyId and SecretAccessKey credentials shared with the CI/CD system
    pipeline_execution_role: Resource
        The IAM role assumed by the pipeline-user to get access to the AWS account and executes the
        CloudFormation stack.
    cloudformation_execution_role: Resource
        The IAM role assumed by the CloudFormation service to executes the CloudFormation stack.
    artifacts_bucket: Resource
        The S3 bucket to hold the SAM build artifacts of the application's CFN template.
    create_image_repository: bool
        A boolean flag that determines whether the user wants to create an ECR image repository or not
    image_repository: ECRImageRepository
        The ECR image repository to hold the image container of lambda functions with Image package-type
    oidc_provider: OidcProvider
        The OIDCProvider to be created/used for assuming the pipeline execution role
    subject_claim: Optional[str]
        The subject claim that will be returned by the OIDC Provider to assume the role

    Methods:
    --------
    did_user_provide_all_required_resources(self) -> bool:
        checks if all of the environment's required resources (pipeline_user, pipeline_execution_role,
        cloudformation_execution_role, artifacts_bucket and image_repository) are provided by the user.
    bootstrap(self, confirm_changeset: bool = True) -> None:
        deploys the CFN template ./stage_resources.yaml to the AWS account identified by aws_profile and
        aws_region member fields. if aws_profile is not provided, it will fallback to  default boto3 credentials'
        resolving. Note that ./stage_resources.yaml template accepts the ARNs of already existing resources(if
        any) as parameters and it will skip the creation of those resources but will use the ARNs to set the proper
        permissions of other missing resources(resources created by the template)
    save_config(self, config_dir: str, filename: str, cmd_names: List[str]):
        save the Artifacts bucket name, ECR image repository URI and ARNs of pipeline_user, pipeline_execution_role and
        cloudformation_execution_role to the "pipelineconfig.toml" file so that it can be auto-filled during
        the `sam pipeline init` command.
    print_resources_summary(self) -> None:
        prints to the screen(console) the ARNs of the created and provided resources.
    _should_create_new_provider(self) -> bool:
        checks if there are any existing OIDC Providers configured in IAM by getting a list of all OIDC Providers
        setup in the account and seeing if the URL provided is in the ARN
    generate_thumbprint(oidc_provider_url) -> Optional[str]:
        retrieves the certificate of the top intermidate cerficate authority that signed the certificate
        used by the external identity provider and then returns the SHA1 hash of it. For more information on
        why the thumbprint is needed and the steps required to obtain it see the following page
        https://docs.aws.amazon.com/IAM/latest/UserGuide/id_roles_providers_create_oidc_verify-thumbprint.html
    """

    def __init__(
        self,
        name: str,
        permissions_provider: Optional[str] = None,
        aws_profile: Optional[str] = None,
        aws_region: Optional[str] = None,
        pipeline_user_arn: Optional[str] = None,
        pipeline_execution_role_arn: Optional[str] = None,
        cloudformation_execution_role_arn: Optional[str] = None,
        artifacts_bucket_arn: Optional[str] = None,
        create_image_repository: bool = False,
        image_repository_arn: Optional[str] = None,
        oidc_provider_url: Optional[str] = None,
        oidc_client_id: Optional[str] = None,
        subject_claim: Optional[str] = None,
        pipeline_oidc_provider: Optional[PipelineOidcProvider] = None,
    ) -> None:
        self.name: str = name
        self.create_new_oidc_provider = False
        self.subject_claim = subject_claim
        self.use_oidc_provider = permissions_provider == OPEN_ID_CONNECT
        self.pipeline_oidc_provider = pipeline_oidc_provider
        self.aws_profile: Optional[str] = aws_profile
        self.aws_region: Optional[str] = aws_region
        self.pipeline_user: IAMUser = IAMUser(arn=pipeline_user_arn, comment="Pipeline IAM user")
        self.pipeline_execution_role: Resource = Resource(
            arn=pipeline_execution_role_arn, comment="Pipeline execution role"
        )
        self.cloudformation_execution_role: Resource = Resource(
            arn=cloudformation_execution_role_arn, comment="CloudFormation execution role"
        )
        self.artifacts_bucket: Resource = Resource(arn=artifacts_bucket_arn, comment="Artifact bucket")
        self.create_image_repository: bool = create_image_repository
        self.image_repository: ECRImageRepository = ECRImageRepository(
            arn=image_repository_arn, comment="ECR image repository"
        )
        self.color = Colored()

        self.oidc_provider: OidcProvider = OidcProvider(
            client_id=oidc_client_id,
            provider_url=oidc_provider_url,
            thumbprint="",
            comment="IAM OIDC Identity Provider",
            arn="",
        )

    def _should_create_new_provider(self, stack_name: str) -> bool:
        """
        Checks if there is an existing Identity Provider in the account already
        whos ARN contains the URL provided by the user.

        OIDC Provider arns are of the following format
        arn:aws:iam:::oidc-provider/api.bitbucket.org/2.0/workspaces//pipelines-config/identity/oidc
        we can check if the URL provided is already in an existing provider to see if a new one should be made
        -------
        """
        if not self.oidc_provider.provider_url:
            return False
        session = boto3.Session(profile_name=self.aws_profile, region_name=self.aws_region)
        iam_client = session.client("iam")
        cfn_client = session.client("cloudformation")
        providers = iam_client.list_open_id_connect_providers()

        url_to_compare = self.oidc_provider.provider_url.replace("https://", "")
        for provider_resource in providers["OpenIDConnectProviderList"]:
            if url_to_compare in provider_resource["Arn"]:
                try:
                    stack_res = cfn_client.describe_stack_resource(
                        StackName=stack_name, LogicalResourceId="OidcProvider"
                    )
                    return url_to_compare in stack_res["StackResourceDetail"]["PhysicalResourceId"]
                except ClientError as ex:
                    if "does not exist" in str(ex):
                        return False
                    raise ex
        return True

    @staticmethod
    def generate_thumbprint(oidc_provider_url: Optional[str]) -> Optional[str]:
        """
        retrieves the certificate of the top intermidate cerficate authority that signed the certificate
        used by the external identity provider and then returns the SHA1 hash of it. For more information on
        why the thumbprint is needed and the steps required to obtain it see the following page
        https://docs.aws.amazon.com/IAM/latest/UserGuide/id_roles_providers_create_oidc_verify-thumbprint.html
        -------
        oidc_provider_url : Optional[str]
            The URL of the OIDC provider that will be used for authentication.
        """
        # Send HTTP GET to retrieve jwks_uri field from openid-configuration document
        oidc_config_url = "{url}/.well-known/openid-configuration".format(url=oidc_provider_url)
        r = requests.get(oidc_config_url, timeout=5)
        jwks_uri = r.json()["jwks_uri"]
        url_for_certificate = urlparse(jwks_uri).hostname

        # Create connection to retrieve certificate
        # Create an IPV4 socket and use TLS for the SSL connection
        address = (url_for_certificate, 443)
        ctx = _get_secure_ssl_context()
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect(address)
        c = SSL.Connection(ctx, s)
        c.set_connect_state()
        # set the servername extension becuase not setting it can cause errors with some sites
        c.set_tlsext_host_name(str.encode(address[0]))

        # If we attempt to get the cert chain without exchanging some traffic it will be empty
        c.sendall(str.encode("HEAD / HTTP/1.0\n\n"))
        peerCertChain = c.get_peer_cert_chain()
        cert = peerCertChain[-1]

        # Dump the certificate in DER/ASN1 format so that its SHA1 hash can be computed
        dumped_cert = crypto.dump_certificate(crypto.FILETYPE_ASN1, cert)
        s.close()

        return hashlib.sha1(dumped_cert).hexdigest()

    def did_user_provide_all_required_resources(self) -> bool:
        """Check if the user provided all of the environment resources or not"""
        return all(resource.is_user_provided for resource in self._get_resources())

    def _get_non_user_provided_resources_msg(self) -> str:
        resource_comments = chain.from_iterable(
            [
                [] if self.pipeline_user.is_user_provided or self.use_oidc_provider else [self.pipeline_user.comment],
                [] if not self.use_oidc_provider else [self.oidc_provider.comment],
                [] if self.pipeline_execution_role.is_user_provided else [self.pipeline_execution_role.comment],
                []
                if self.cloudformation_execution_role.is_user_provided
                else [self.cloudformation_execution_role.comment],
                [] if self.artifacts_bucket.is_user_provided else [self.artifacts_bucket.comment],
                []
                if self.image_repository.is_user_provided or not self.create_image_repository
                else [self.image_repository.comment],
            ]
        )
        return "\n".join([f"\t- {comment}" for comment in resource_comments])

    def bootstrap(self, confirm_changeset: bool = True) -> bool:
        """
        Deploys the CFN template(./stage_resources.yaml) which deploys:
            * Pipeline IAM User or IAM OIDC Identity Provider
            * Pipeline execution IAM role
            * CloudFormation execution IAM role
            * Artifacts' S3 Bucket
            * ECR image repository
        to the AWS account associated with the given environment. It will not redeploy the stack if already exists.
        This CFN template accepts the ARNs of the resources as parameters and will not create a resource if already
        provided, this way we can conditionally create a resource only if the user didn't provide it

        THIS METHOD UPDATES THE STATE OF THE CALLING INSTANCE(self) IT WILL SET THE VALUES OF THE RESOURCES ATTRIBUTES

        Parameters
        ----------
        confirm_changeset: bool
            if set to false, the stage_resources.yaml CFN template will directly be deployed, otherwise,
            the user will be prompted for confirmation

        Returns True if bootstrapped, otherwise False
        """

        if self.did_user_provide_all_required_resources():
            click.secho(
                self.color.yellow(
                    f"\nAll required resources for the {self.name} configuration exist, skipping creation."
                )
            )
            return True

        missing_resources_msg: str = self._get_non_user_provided_resources_msg()
        click.echo(
            f"This will create the following required resources for the '{self.name}' configuration: \n"
            f"{missing_resources_msg}"
        )
        if confirm_changeset:
            confirmed: bool = click.confirm("Should we proceed with the creation?")
            if not confirmed:
                click.secho(self.color.red("Canceling pipeline bootstrap creation."))
                return False

        stack_name = self._get_stack_name()

        if self.use_oidc_provider:
            self.create_new_oidc_provider = self._should_create_new_provider(stack_name)
            if self.create_new_oidc_provider:
                self.oidc_provider.thumbprint = self.generate_thumbprint(self.oidc_provider.provider_url)

        environment_resources_template_body = Stage._read_template(STAGE_RESOURCES_CFN_TEMPLATE)
        output: StackOutput = update_stack(
            stack_name=stack_name,
            region=self.aws_region,
            profile=self.aws_profile,
            template_body=environment_resources_template_body,
            parameter_overrides={
                "PipelineUserArn": self.pipeline_user.arn or "",
                "PipelineExecutionRoleArn": self.pipeline_execution_role.arn or "",
                "CloudFormationExecutionRoleArn": self.cloudformation_execution_role.arn or "",
                "ArtifactsBucketArn": self.artifacts_bucket.arn or "",
                "CreateImageRepository": "true" if self.create_image_repository else "false",
                "ImageRepositoryArn": self.image_repository.arn or "",
                "IdentityProviderThumbprint": self.oidc_provider.thumbprint or "",
                "OidcClientId": self.oidc_provider.client_id or "",
                "OidcProviderUrl": self.oidc_provider.provider_url or "",
                "UseOidcProvider": "true" if self.use_oidc_provider else "false",
                "SubjectClaim": self.subject_claim or "",
                "CreateNewOidcProvider": "true" if self.create_new_oidc_provider else "false",
            },
        )

        if not self.use_oidc_provider:
            pipeline_user_secret_sm_id = output.get("PipelineUserSecretKey")

            self.pipeline_user.arn = output.get("PipelineUser")
            if pipeline_user_secret_sm_id:
                (
                    self.pipeline_user.access_key_id,
                    self.pipeline_user.secret_access_key,
                ) = Stage._get_pipeline_user_secret_pair(pipeline_user_secret_sm_id, self.aws_profile, self.aws_region)
        self.pipeline_execution_role.arn = output.get("PipelineExecutionRole")
        self.cloudformation_execution_role.arn = output.get("CloudFormationExecutionRole")
        self.artifacts_bucket.arn = output.get("ArtifactsBucket")
        self.image_repository.arn = output.get("ImageRepository")
        return True

    @staticmethod
    def _get_pipeline_user_secret_pair(
        secret_manager_arn: str, profile: Optional[str], region: Optional[str]
    ) -> Tuple[str, str]:
        """
        Helper method to fetch pipeline user's AWS Credentials from secrets manager.
        SecretString need to be in following JSON format:
        {
          "aws_access_key_id": "AWSSECRETACCESSKEY123",
          "aws_secret_access_key": "mYSuperSecretDummyKey"
        }
        Parameters
        ----------
        secret_manager_arn:
            ARN of secret manager entry which holds pipeline user key.
        profile:
            The named AWS profile (in user's machine) of the AWS account to deploy this environment to.
        region:
            The AWS region to deploy this environment to.

        Returns tuple of aws_access_key_id and aws_secret_access_key.

        """
        session = boto3.Session(profile_name=profile, region_name=region if region else None)
        secrets_manager_client = session.client("secretsmanager")
        response = secrets_manager_client.get_secret_value(SecretId=secret_manager_arn)
        secret_string = response["SecretString"]
        secret_json = json.loads(secret_string)
        return secret_json["aws_access_key_id"], secret_json["aws_secret_access_key"]

    @staticmethod
    def _read_template(template_file_name: str) -> str:
        template_path: str = os.path.join(CFN_TEMPLATE_PATH, template_file_name)
        with open(template_path, "r", encoding="utf-8") as fp:
            template_body = fp.read()
        return template_body

    def save_config(self, config_dir: str, filename: str, cmd_names: List[str]) -> None:
        """
        save the Artifacts bucket name, ECR image repository URI and ARNs of pipeline_user, pipeline_execution_role and
        cloudformation_execution_role to the given filename and directory.

        Parameters
        ----------
        config_dir: str
            the directory of the toml file to save to
        filename: str
            the name of the toml file to save to
        cmd_names: List[str]
            nested command name to scope the saved configs to inside the toml file

        Raises
        ------
        ValueError: if the artifacts_bucket or ImageRepository ARNs are invalid
        """

        samconfig: SamConfig = SamConfig(config_dir=config_dir, filename=filename)

        if self.pipeline_user.arn:
            samconfig.put(cmd_names=cmd_names, section="parameters", key=PIPELINE_USER, value=self.pipeline_user.arn)
            samconfig.put(cmd_names=cmd_names, section="parameters", key=PERMISSIONS_PROVIDER, value="AWS IAM")
        if self.use_oidc_provider and self.pipeline_oidc_provider:
            self.pipeline_oidc_provider.save_values(cmd_names=cmd_names, section="parameters", samconfig=samconfig)

        # Computing Artifacts bucket name and ECR image repository URL may through an exception if the ARNs are wrong
        # Let's swallow such an exception to be able to save the remaining resources
        try:
            artifacts_bucket_name: Optional[str] = self.artifacts_bucket.name()
        except ValueError:
            artifacts_bucket_name = ""
        try:
            image_repository_uri: Optional[str] = self.image_repository.get_uri() or ""
        except ValueError:
            image_repository_uri = ""

        environment_specific_configs: Dict[str, Optional[str]] = {
            PIPELINE_EXECUTION_ROLE: self.pipeline_execution_role.arn,
            CLOUDFORMATION_EXECUTION_ROLE: self.cloudformation_execution_role.arn,
            ARTIFACTS_BUCKET: artifacts_bucket_name,
            # even image repository can be None, we want to save it as empty string
            # so that pipeline init command can pick it up
            ECR_IMAGE_REPOSITORY: image_repository_uri,
            REGION: self.aws_region,
        }

        for key, value in environment_specific_configs.items():
            if value is not None:
                samconfig.put(
                    cmd_names=cmd_names,
                    section="parameters",
                    key=key,
                    value=value,
                    env=self.name,
                )

        samconfig.flush()

    def save_config_safe(self, config_dir: str, filename: str, cmd_names: List[str]) -> None:
        """
        A safe version of save_config method that doesn't raise any exception
        """
        try:
            self.save_config(config_dir, filename, cmd_names)
        except Exception:
            pass

    def _get_resources(self) -> List[Resource]:
        resources = [
            self.pipeline_execution_role,
            self.cloudformation_execution_role,
            self.artifacts_bucket,
        ]
        if self.use_oidc_provider:
            resources.append(self.oidc_provider)
        else:
            resources.append(self.pipeline_user)
        if self.create_image_repository or self.image_repository.arn:  # Image Repository is optional
            resources.append(self.image_repository)
        return resources

    def print_resources_summary(self) -> None:
        """prints to the screen(console) the ARNs of the created and provided resources."""

        provided_resources = []
        created_resources = []
        for resource in self._get_resources():
            if resource.is_user_provided:
                provided_resources.append(resource)
            else:
                created_resources.append(resource)

        if created_resources:
            click.secho(self.color.green("The following resources were created in your account:"))
            for resource in created_resources:
                click.secho(self.color.green(f"\t- {resource.comment}"))

        if not self.pipeline_user.is_user_provided and not self.use_oidc_provider:
            click.secho(self.color.green("Pipeline IAM user credential:"))
            click.secho(self.color.green(f"\tAWS_ACCESS_KEY_ID: {self.pipeline_user.access_key_id}"))
            click.secho(self.color.green(f"\tAWS_SECRET_ACCESS_KEY: {self.pipeline_user.secret_access_key}"))

    def _get_stack_name(self) -> str:
        sanitized_stage_configuration_name: str = re.sub("[^0-9a-zA-Z]+", "-", self.name)
        return f"{STACK_NAME_PREFIX}-{sanitized_stage_configuration_name}-{STAGE_RESOURCES_STACK_NAME_SUFFIX}"



"""
Interfaces and generic implementations for observability events (like CW logs)
"""
import logging
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Dict, Generic, List, Optional, Sequence, TypeVar, Union

from samcli.lib.utils.async_utils import AsyncContext

LOG = logging.getLogger(__name__)

# Generic type for the internal observability event
InternalEventType = TypeVar("InternalEventType")


class ObservabilityEvent(Generic[InternalEventType]):
    """
    Generic class that represents observability event
    This keeps some common fields for filtering or sorting later on
    """

    def __init__(self, event: InternalEventType, timestamp: int, resource_name: Optional[str] = None):
        """
        Parameters
        ----------
        event : EventType
            Actual event object. This can be any type with generic definition (dict, str etc.)
        timestamp : int
            Timestamp of the event
        resource_name : Optional[str]
            Resource name related to this event. This is optional since not all events is connected to a single resource
        """
        self.event = event
        self.timestamp = timestamp
        self.resource_name = resource_name


# Generic type for identifying different ObservabilityEvent
ObservabilityEventType = TypeVar("ObservabilityEventType", bound=ObservabilityEvent)


class ObservabilityPuller(ABC):
    """
    Interface definition for pulling observability information.
    """

    # used to cancel indefinitely running processes (eg: tail)
    cancelled: bool = False

    @abstractmethod
    def tail(self, start_time: Optional[datetime] = None, filter_pattern: Optional[str] = None):
        """
        Parameters
        ----------
        start_time : Optional[datetime]
            Optional parameter to tail information from earlier time
        filter_pattern :  Optional[str]
            Optional parameter to filter events with given string
        """

    @abstractmethod
    def load_time_period(
        self,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        filter_pattern: Optional[str] = None,
    ):
        """
        Parameters
        ----------
        start_time : Optional[datetime]
            Optional parameter to load events from certain date time
        end_time :  Optional[datetime]
            Optional parameter to load events until certain date time
        filter_pattern : Optional[str]
            Optional parameter to filter events with given string
        """

    @abstractmethod
    def load_events(self, event_ids: Union[List[Any], Dict]):
        """
        This method will load specific events which is given by the event_ids parameter

        Parameters
        ----------
        event_ids : List[str] or Dict
            List of event ids that will be pulled
        """

    def stop_tailing(self):
        self.cancelled = True


# pylint: disable=fixme
# fixme add ABC parent class back once we bump the pylint to a version 2.8.2 or higher
class ObservabilityEventMapper(Generic[ObservabilityEventType]):
    """
    Interface definition to map/change any event to another object
    This could be used by highlighting certain parts or formatting events before logging into console
    """

    @abstractmethod
    def map(self, event: ObservabilityEventType) -> Any:
        """
        Parameters
        ----------
        event : ObservabilityEventType
            Event object that will be mapped/converted to another event or any object

        Returns
        -------
        Any
            Return converted type
        """


class ObservabilityEventConsumer(Generic[ObservabilityEventType]):
    """
    Consumer interface, which will consume any event.
    An example is to output event into console.
    """

    @abstractmethod
    def consume(self, event: ObservabilityEventType):
        """
        Parameters
        ----------
        event : ObservabilityEvent
            Event that will be consumed
        """


class ObservabilityEventConsumerDecorator(ObservabilityEventConsumer):
    """
    A decorator implementation for consumer, which can have mappers and decorated consumer within.
    Rather than the normal implementation, this will process the events through mappers which is been
    provided, and then pass them to actual consumer
    """

    def __init__(self, mappers: List[ObservabilityEventMapper], consumer: ObservabilityEventConsumer):
        """
        Parameters
        ----------
        mappers : List[ObservabilityEventMapper]
            List of event mappers which will be used to process events before passing to consumer
        consumer : ObservabilityEventConsumer
            Actual consumer which will handle the events after they are processed by mappers
        """
        super().__init__()
        self._mappers = mappers
        self._consumer = consumer

    def consume(self, event: ObservabilityEvent):
        """
        See Also ObservabilityEventConsumerDecorator and ObservabilityEventConsumer
        """
        for mapper in self._mappers:
            LOG.debug("Calling mapper (%s) for event (%s)", mapper, event)
            event = mapper.map(event)
        LOG.debug("Calling consumer (%s) for event (%s)", self._consumer, event)
        self._consumer.consume(event)


class ObservabilityCombinedPuller(ObservabilityPuller):
    """
    A decorator class which will contain multiple ObservabilityPuller instance and pull information from each of them
    """

    def __init__(self, pullers: Sequence[ObservabilityPuller]):
        """
        Parameters
        ----------
        pullers : List[ObservabilityPuller]
            List of pullers which will be managed by this class
        """
        self._pullers = pullers

    def tail(self, start_time: Optional[datetime] = None, filter_pattern: Optional[str] = None):
        """
        Implementation of ObservabilityPuller.tail method with AsyncContext.
        It will create tasks by calling tail methods of all given pullers, and execute them in async
        """
        async_context = AsyncContext()
        for puller in self._pullers:
            LOG.debug("Adding task 'tail' for puller (%s)", puller)
            async_context.add_async_task(puller.tail, start_time, filter_pattern)
        LOG.debug("Running all 'tail' tasks in parallel")
        try:
            async_context.run_async()
        except KeyboardInterrupt:
            LOG.info(" CTRL+C received, cancelling...")
            self.stop_tailing()

    def load_time_period(
        self,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        filter_pattern: Optional[str] = None,
    ):
        """
        Implementation of ObservabilityPuller.load_time_period method with AsyncContext.
        It will create tasks by calling load_time_period methods of all given pullers, and execute them in async
        """
        async_context = AsyncContext()
        for puller in self._pullers:
            LOG.debug("Adding task 'load_time_period' for puller (%s)", puller)
            async_context.add_async_task(puller.load_time_period, start_time, end_time, filter_pattern)
        LOG.debug("Running all 'load_time_period' tasks in parallel")
        async_context.run_async()

    def load_events(self, event_ids: Union[List[Any], Dict]):
        """
        Implementation of ObservabilityPuller.load_events method with AsyncContext.
        It will create tasks by calling load_events methods of all given pullers, and execute them in async
        """
        async_context = AsyncContext()
        for puller in self._pullers:
            LOG.debug("Adding task 'load_events' for puller (%s)", puller)
            async_context.add_async_task(puller.load_events, event_ids)
        LOG.debug("Running all 'load_time_period' tasks in parallel")
        async_context.run_async()

    def stop_tailing(self):
        # if ObservabilityCombinedPuller A is a child puller in other ObservabilityCombinedPuller B, make sure A's child
        # pullers stop as well when B stops.
        for puller in self._pullers:
            puller.stop_tailing()



"""
Utils for CDK-based projects
"""

import os
from typing import Dict

CDK_METADATA_TYPE_VALUE = "AWS::CDK::Metadata"
CDK_PATH_METADATA_KEY = "aws:cdk:path"


def is_cdk_project(template: Dict) -> bool:
    """
    Check if the CFN template was produced from a CDK project

    Parameters
    ----------
    template:
        CFN template to look through

    """
    resources = template.get("Resources", {})
    cdk_project_conditions = any(
        [
            _resource_level_metadata_exists(resources),
            _cdk_path_metadata_exists(resources),
            _relevant_cdk_files_are_present(),
        ]
    )
    return cdk_project_conditions


def _resource_level_metadata_exists(resources: Dict) -> bool:
    """
    Check if there is a AWS::CDK::Metadata resource is in the resources

    Parameters
    ----------
    resources:
        Dict of resources to look through

    """
    for _, resource in resources.items():
        if resource.get("Type", "") == CDK_METADATA_TYPE_VALUE:
            return True
    return False


def _cdk_path_metadata_exists(resources: Dict) -> bool:
    """
    Check if there is an aws:cdk:path property in the metadata of any of the resources

    Parameters
    ----------
    resources:
        Dict of resources to look through

    """
    for _, resource in resources.items():
        metadata = resource.get("Metadata", {})
        if metadata and CDK_PATH_METADATA_KEY in metadata:
            return True
    return False


def _relevant_cdk_files_are_present() -> bool:
    """
    Check if there are CDK files in the cwd
    """
    relevant_cdk_files = ["manifest.json", "tree.json", "cdk.json"]
    # In case a customer runs `cdk synth --no-staging > template.yaml` the template
    # will be in the project root and it's entirely possible to find the cdk.json there
    project_files = os.listdir(os.getcwd())
    return any(cdk_file in project_files for cdk_file in relevant_cdk_files)



"""Exceptions related to sync functionalities"""
from typing import TYPE_CHECKING, Dict, Optional

if TYPE_CHECKING:  # pragma: no cover
    from samcli.lib.sync.sync_flow import SyncFlow


class SyncFlowException(Exception):
    """Exception wrapper for exceptions raised in SyncFlows"""

    _sync_flow: "SyncFlow"
    _exception: Exception

    def __init__(self, sync_flow: "SyncFlow", exception: Exception):
        """
        Parameters
        ----------
        sync_flow : SyncFlow
            SyncFlow that raised the exception
        exception : Exception
            exception raised
        """
        super().__init__(f"SyncFlow Exception for {sync_flow.log_name}")
        self._sync_flow = sync_flow
        self._exception = exception

    @property
    def sync_flow(self) -> "SyncFlow":
        return self._sync_flow

    @property
    def exception(self) -> Exception:
        return self._exception


class InfraSyncRequiredError(Exception):
    """Exception used if SyncFlow cannot handle the sync and an infra sync is required"""

    _resource_identifier: Optional[str]
    _reason: Optional[str]

    def __init__(self, resource_identifier: Optional[str] = None, reason: Optional[str] = ""):
        """
        Parameters
        ----------
        resource_identifier : str
            Logical resource identifier
        reason : str
            Reason for requiring infra sync
        """
        super().__init__(f"{resource_identifier} cannot be code synced.")
        self._resource_identifier = resource_identifier
        self._reason = reason

    @property
    def resource_identifier(self) -> Optional[str]:
        """
        Returns
        -------
        str
            Resource identifier of the resource that does not have a remote/physical counterpart
        """
        return self._resource_identifier

    @property
    def reason(self) -> Optional[str]:
        """
        Returns
        -------
        str
            Reason to why the SyncFlow cannot sync the resource
        """
        return self._reason


class MissingPhysicalResourceError(Exception):
    """Exception used for not having a remote/physical counterpart for a local stack resource"""

    _resource_identifier: Optional[str]
    _physical_resource_mapping: Optional[Dict[str, str]]

    def __init__(
        self, resource_identifier: Optional[str] = None, physical_resource_mapping: Optional[Dict[str, str]] = None
    ):
        """
        Parameters
        ----------
        resource_identifier : str
            Logical resource identifier
        physical_resource_mapping: Dict[str, str]
            Current mapping between logical and physical IDs
        """
        super().__init__(f"{resource_identifier} is not found in remote.")
        self._resource_identifier = resource_identifier
        self._physical_resource_mapping = physical_resource_mapping

    @property
    def resource_identifier(self) -> Optional[str]:
        """
        Returns
        -------
        str
            Resource identifier of the resource that does not have a remote/physical counterpart
        """
        return self._resource_identifier

    @property
    def physical_resource_mapping(self) -> Optional[Dict[str, str]]:
        """
        Returns
        -------
        Optional[Dict[str, str]]
            Physical ID mapping for resources when the excecption was raised
        """
        return self._physical_resource_mapping


class NoLayerVersionsFoundError(Exception):
    """This is used when we try to list all versions for layer, but we found none"""

    _layer_name_arn: str

    def __init__(self, layer_name_arn: str):
        """
        Parameters
        ----------
        layer_name_arn : str
            Layer ARN without version info at the end of it
        """
        super().__init__(f"{layer_name_arn} doesn't have any versions in remote.")
        self._layer_name_arn = layer_name_arn

    @property
    def layer_name_arn(self) -> str:
        """
        Returns
        -------
        str
            Layer ARN without version info at the end of it
        """
        return self._layer_name_arn


class MissingLockException(Exception):
    """Exception for not having an associated lock to be used."""


class MissingFunctionBuildDefinition(Exception):
    """This is used when no build definition found for particular function"""

    _function_logical_id: str

    def __init__(self, function_logical_id: str):
        super().__init__(f"Build definition for {function_logical_id} can't be found")
        self._function_logical_id = function_logical_id

    @property
    def function_logical_id(self) -> str:
        return self._function_logical_id


class InvalidRuntimeDefinitionForFunction(Exception):
    """This is used when no Runtime information is defined for a function resource"""

    _function_logical_id: str

    def __init__(self, function_logical_id):
        super().__init__(f"Invalid Runtime definition for {function_logical_id}")
        self._function_logical_id = function_logical_id

    @property
    def function_logical_id(self):
        return self._function_logical_id



"""
    Companion stack template builder
"""
from typing import Dict, cast

from samcli.lib.bootstrap.companion_stack.data_types import CompanionStack, ECRRepo
from samcli.lib.bootstrap.stack_builder import AbstractStackBuilder


class CompanionStackBuilder(AbstractStackBuilder):
    """
    CFN template builder for the companion stack
    """

    _parent_stack_name: str
    _companion_stack: CompanionStack
    _repo_mapping: Dict[str, ECRRepo]

    def __init__(self, companion_stack: CompanionStack) -> None:
        super().__init__("AWS SAM CLI Managed ECR Repo Stack")
        self._companion_stack = companion_stack
        self._repo_mapping: Dict[str, ECRRepo] = dict()
        self.add_metadata("CompanionStackname", self._companion_stack.stack_name)

    def add_function(self, function_logical_id: str) -> None:
        """
        Add an ECR repo associated with the function to the companion stack template
        """
        self._repo_mapping[function_logical_id] = ECRRepo(self._companion_stack, function_logical_id)

    def clear_functions(self) -> None:
        """
        Remove all functions that need ECR repos
        """
        self._repo_mapping = dict()

    def build(self) -> str:
        """
        Build companion stack CFN template with current functions
        Returns
        -------
        str
            CFN template for companions stack
        """
        for _, ecr_repo in self._repo_mapping.items():
            self.add_resource(cast(str, ecr_repo.logical_id), self._build_repo_dict(ecr_repo))
            self.add_output(cast(str, ecr_repo.output_logical_id), CompanionStackBuilder._build_output_dict(ecr_repo))

        return super().build()

    def _build_repo_dict(self, repo: ECRRepo) -> Dict:
        """
        Build a single ECR repo resource dictionary

        Parameters
        ----------
        repo
            ECR repo that will be turned into CFN resource

        Returns
        -------
        dict
            ECR repo resource dictionary
        """
        return {
            "Type": "AWS::ECR::Repository",
            "Properties": {
                "RepositoryName": repo.physical_id,
                "Tags": [
                    {"Key": "ManagedStackSource", "Value": "AwsSamCli"},
                    {"Key": "AwsSamCliCompanionStack", "Value": self._companion_stack.stack_name},
                ],
                "RepositoryPolicyText": {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Sid": "AllowLambdaSLR",
                            "Effect": "Allow",
                            "Principal": {"Service": ["lambda.amazonaws.com"]},
                            "Action": ["ecr:GetDownloadUrlForLayer", "ecr:GetRepositoryPolicy", "ecr:BatchGetImage"],
                        }
                    ],
                },
            },
        }

    @staticmethod
    def _build_output_dict(repo: ECRRepo) -> str:
        """
        Build a single ECR repo output resource dictionary

        Parameters
        ----------
        repo
            ECR repo that will be turned into CFN output resource

        Returns
        -------
        dict
            ECR repo output resource dictionary
        """
        return f"!Sub ${{AWS::AccountId}}.dkr.ecr.${{AWS::Region}}.${{AWS::URLSuffix}}/${{{repo.logical_id}}}"

    @property
    def repo_mapping(self) -> Dict[str, ECRRepo]:
        """
        Repo mapping dictionary with key as function logical ID and value as ECRRepo object
        """
        return self._repo_mapping



"""
Sets up the cli for resources
"""

import click

from samcli.cli.cli_config_file import TomlProvider, configuration_option
from samcli.cli.main import aws_creds_options, common_options, pass_context, print_cmdline_args
from samcli.commands._utils.command_exception_handler import command_exception_handler
from samcli.commands._utils.options import template_option_without_build
from samcli.commands.list.cli_common.options import output_option, stack_name_not_provided_message, stack_name_option
from samcli.lib.telemetry.metric import track_command
from samcli.lib.utils.version_checker import check_newer_version

HELP_TEXT = """
Get a list of resources that will be deployed to CloudFormation.\n
If a stack name is provided, the corresponding physical IDs of each
resource will be mapped to the logical ID of each resource.
"""


@click.command(name="resources", help=HELP_TEXT)
@configuration_option(provider=TomlProvider(section="parameters"))
@stack_name_option
@output_option
@template_option_without_build
@aws_creds_options
@common_options
@pass_context
@track_command
@check_newer_version
@print_cmdline_args
@command_exception_handler
def cli(self, stack_name, output, template_file, config_file, config_env):
    """
    `sam list resources` command entry point
    """

    do_cli(stack_name=stack_name, output=output, region=self.region, profile=self.profile, template_file=template_file)


def do_cli(stack_name, output, region, profile, template_file):
    """
    Implementation of the ``cli`` method
    """
    from samcli.commands.list.resources.resources_context import ResourcesContext

    with ResourcesContext(
        stack_name=stack_name, output=output, region=region, profile=profile, template_file=template_file
    ) as resources_context:
        if not stack_name:
            stack_name_not_provided_message()
        resources_context.run()



"""
Implementation of Local Lambda runner
"""

import logging
import os
from typing import Any, Dict, Optional, cast

import boto3
from botocore.credentials import Credentials

from samcli.commands.local.lib.debug_context import DebugContext
from samcli.commands.local.lib.exceptions import (
    InvalidIntermediateImageError,
    NoPrivilegeException,
    OverridesNotWellDefinedError,
    UnsupportedInlineCodeError,
)
from samcli.lib.providers.provider import Function
from samcli.lib.providers.sam_function_provider import SamFunctionProvider
from samcli.lib.utils.architecture import validate_architecture_runtime
from samcli.lib.utils.codeuri import resolve_code_path
from samcli.lib.utils.packagetype import IMAGE, ZIP
from samcli.lib.utils.stream_writer import StreamWriter
from samcli.local.docker.container import ContainerConnectionTimeoutException, ContainerResponseException
from samcli.local.lambdafn.config import FunctionConfig
from samcli.local.lambdafn.env_vars import EnvironmentVariables
from samcli.local.lambdafn.exceptions import FunctionNotFound
from samcli.local.lambdafn.runtime import LambdaRuntime

LOG = logging.getLogger(__name__)


class LocalLambdaRunner:
    """
    Runs Lambda functions locally. This class is a wrapper around the `samcli.local` library which takes care
    of actually running the function on a Docker container.
    """

    MAX_DEBUG_TIMEOUT = 36000  # 10 hours in seconds
    WIN_ERROR_CODE = 1314

    def __init__(
        self,
        local_runtime: LambdaRuntime,
        function_provider: SamFunctionProvider,
        cwd: str,
        aws_profile: Optional[str] = None,
        aws_region: Optional[str] = None,
        env_vars_values: Optional[Dict[Any, Any]] = None,
        debug_context: Optional[DebugContext] = None,
        container_host: Optional[str] = None,
        container_host_interface: Optional[str] = None,
    ) -> None:
        """
        Initializes the class

        :param samcli.local.lambdafn.runtime.LambdaRuntime local_runtime: Lambda runtime capable of running a function
        :param samcli.commands.local.lib.provider.FunctionProvider function_provider: Provider that can return a
            Lambda function
        :param string cwd: Current working directory. We will resolve all function CodeURIs relative to this directory.
        :param string aws_profile: Optional. Name of the profile to fetch AWS credentials from.
        :param string aws_region: Optional. AWS Region to use.
        :param dict env_vars_values: Optional. Dictionary containing values of environment variables.
        :param DebugContext debug_context: Optional. Debug context for the function (includes port, args, and path).
        :param string container_host: Optional. Host of locally emulated Lambda container
        :param string container_host_interface: Optional. Interface that Docker host binds ports to
        """

        self.local_runtime = local_runtime
        self.provider = function_provider
        self.cwd = cwd
        self.aws_profile = aws_profile
        self.aws_region = aws_region
        self.env_vars_values = env_vars_values or {}
        self.debug_context = debug_context
        self._boto3_session_creds: Optional[Credentials] = None
        self._boto3_region: Optional[str] = None
        self.container_host = container_host
        self.container_host_interface = container_host_interface

    def invoke(
        self,
        function_identifier: str,
        event: str,
        stdout: Optional[StreamWriter] = None,
        stderr: Optional[StreamWriter] = None,
    ) -> None:
        """
        Find the Lambda function with given name and invoke it. Pass the given event to the function and return
        response through the given streams.

        This function will block until either the function completes or times out.

        Parameters
        ----------
        function_identifier str
            Identifier of the Lambda function to invoke, it can be logicalID, function name or full path
        event str
            Event data passed to the function. Must be a valid JSON String.
        stdout samcli.lib.utils.stream_writer.StreamWriter
            Stream writer to write the output of the Lambda function to.
        stderr samcli.lib.utils.stream_writer.StreamWriter
            Stream writer to write the Lambda runtime logs to.

        Raises
        ------
        FunctionNotfound
            When we cannot find a function with the given name
        """

        # Generate the correct configuration based on given inputs
        function = self.provider.get(function_identifier)

        if not function:
            all_function_full_paths = [f.full_path for f in self.provider.get_all()]
            available_function_message = "{} not found. Possible options in your template: {}".format(
                function_identifier, all_function_full_paths
            )
            LOG.info(available_function_message)
            raise FunctionNotFound("Unable to find a Function with name '{}'".format(function_identifier))

        LOG.debug("Found one Lambda function with name '%s'", function_identifier)
        if function.packagetype == ZIP:
            if function.inlinecode:
                raise UnsupportedInlineCodeError(
                    "Inline code is not supported for sam local commands."
                    f" Please write your code in a separate file for the function {function.function_id}."
                )
            LOG.info("Invoking %s (%s)", function.handler, function.runtime)
        elif function.packagetype == IMAGE:
            if not function.imageuri:
                raise InvalidIntermediateImageError(
                    f"ImageUri not provided for Function: {function_identifier} of PackageType: {function.packagetype}"
                )
            LOG.info("Invoking Container created from %s", function.imageuri)

        validate_architecture_runtime(function)

        config = self.get_invoke_config(function)

        # Invoke the function
        try:
            self.local_runtime.invoke(
                config,
                event,
                debug_context=self.debug_context,
                stdout=stdout,
                stderr=stderr,
                container_host=self.container_host,
                container_host_interface=self.container_host_interface,
            )
        except ContainerResponseException:
            # NOTE(sriram-mv): This should still result in a exit code zero to avoid regressions.
            LOG.info("No response from invoke container for %s", function.name)
        except ContainerConnectionTimeoutException as e:
            # NOTE: Exit code of zero here as well to match the behaviour above (ContainerResponseException
            # having exit code of zero) because previously when it timed out or exhausted retries while
            # trying to connect to the socket for Docker it would throw ContainerResponseException but now it's this.
            LOG.info(str(e))
        except OSError as os_error:
            if getattr(os_error, "winerror", None) == self.WIN_ERROR_CODE:
                raise NoPrivilegeException(
                    "Administrator, Windows Developer Mode, "
                    "or SeCreateSymbolicLinkPrivilege is required to create symbolic link for files: {}, {}".format(
                        os_error.filename, os_error.filename2
                    )
                ) from os_error

            raise

    def is_debugging(self) -> bool:
        """
        Are we debugging the invoke?

        Returns
        -------
        bool
            True, if we are debugging the invoke ie. the Docker container will break into the debugger and wait for
            attach
        """
        return bool(self.debug_context)

    def get_invoke_config(self, function: Function) -> FunctionConfig:
        """
        Returns invoke configuration to pass to Lambda Runtime to invoke the given function

        :param samcli.commands.local.lib.provider.Function function: Lambda function to generate the configuration for
        :return samcli.local.lambdafn.config.FunctionConfig: Function configuration to pass to Lambda runtime
        """

        env_vars = self._make_env_vars(function)
        code_abs_path = None
        if function.packagetype == ZIP:
            code_abs_path = resolve_code_path(self.cwd, function.codeuri)
            LOG.debug("Resolved absolute path to code is %s", code_abs_path)

        function_timeout = function.timeout

        # The Runtime container handles timeout inside the container. When debugging with short timeouts, this can
        # cause the container execution to stop. When in debug mode, we set the timeout in the container to a max 10
        # hours. This will ensure the container doesn't unexpectedly stop while debugging function code
        if self.is_debugging():
            function_timeout = self.MAX_DEBUG_TIMEOUT

        return FunctionConfig(
            name=function.name,
            full_path=function.full_path,
            runtime=function.runtime,
            handler=function.handler,
            imageuri=function.imageuri,
            imageconfig=function.imageconfig,
            packagetype=function.packagetype,
            code_abs_path=code_abs_path,
            layers=function.layers,
            architecture=function.architecture,
            memory=function.memory,
            timeout=function_timeout,
            env_vars=env_vars,
            runtime_management_config=function.runtime_management_config,
        )

    def _make_env_vars(self, function: Function) -> EnvironmentVariables:
        """Returns the environment variables configuration for this function

        Priority order for environment variables (high to low):
        1. Function specific env vars from json file
        2. Global env vars from json file

        Parameters
        ----------
        function : samcli.commands.local.lib.provider.Function
            Lambda function to generate the configuration for

        Returns
        -------
        samcli.local.lambdafn.env_vars.EnvironmentVariables
            Environment variable configuration for this function

        Raises
        ------
        samcli.commands.local.lib.exceptions.OverridesNotWellDefinedError
            If the environment dict is in the wrong format to process environment vars

        """

        function_id = function.function_id
        logical_id = function.name
        function_name = function.functionname
        full_path = function.full_path

        variables = None
        if isinstance(function.environment, dict) and "Variables" in function.environment:
            variables = function.environment["Variables"]
        else:
            LOG.debug("No environment variables found for function '%s'", logical_id)

        # This could either be in standard format, or a CloudFormation parameter file format, or mix of both.
        #
        # Standard format is {FunctionName: {key:value}, FunctionName: {key:value}}
        # CloudFormation parameter file is {"Parameters": {key:value}}
        # Mixed format is {FunctionName: {key:value}, "Parameters": {key:value}}

        for env_var_value in self.env_vars_values.values():
            if not isinstance(env_var_value, dict):
                reason = "Environment variables {} in incorrect format".format(env_var_value)
                LOG.debug(reason)
                raise OverridesNotWellDefinedError(reason)

        overrides = {}
        # environment variables for specific resources take precedence over
        # the single environment variable for all resources
        if "Parameters" in self.env_vars_values:
            LOG.debug("Environment variables data found in the CloudFormation parameter file format")
            # CloudFormation parameter file format
            parameter_result = self.env_vars_values.get("Parameters", {})
            overrides.update(parameter_result)

        # Precedence: logical_id -> function_id -> function name -> full_path, customer can use any of them
        fn_file_env_vars = (
            self.env_vars_values.get(logical_id, None)
            or self.env_vars_values.get(function_id, None)
            or self.env_vars_values.get(function_name, None)
            or self.env_vars_values.get(full_path, None)
        )
        if fn_file_env_vars:
            # Standard format
            LOG.debug("Environment variables data found for specific function in standard format")
            overrides.update(fn_file_env_vars)

        shell_env = os.environ
        aws_creds = self.get_aws_creds()

        return EnvironmentVariables(
            function.name,
            function.memory,
            function.timeout,
            function.handler,
            variables=variables,
            shell_env_values=shell_env,
            override_values=overrides,
            aws_creds=aws_creds,
        )  # EnvironmentVariables is not yet annotated with type hints, disable mypy check for now. type: ignore

    def _get_session_creds(self) -> Optional[Credentials]:
        if self._boto3_session_creds is None:
            # to pass command line arguments for region & profile to setup boto3 default session
            LOG.debug("Loading AWS credentials from session with profile '%s'", self.aws_profile)
            # The signature of boto3.session.Session defines the default values of profile_name and region_name
            # so they should be Optional. But mypy follows its docstring which does not have "Optional."
            # Here we trick mypy thinking they are both str rather than Optional[str].
            session = boto3.session.Session(
                profile_name=cast(str, self.aws_profile), region_name=cast(str, self.aws_region)
            )

            # check for region_name in session and cache
            if hasattr(session, "region_name") and session.region_name:
                self._boto3_region = session.region_name

            # don't set cached session creds if there is not a session
            if session:
                self._boto3_session_creds = session.get_credentials()

        return self._boto3_session_creds

    def get_aws_creds(self) -> Dict[str, str]:
        """
        Returns AWS credentials obtained from the shell environment or given profile

        :return dict: A dictionary containing credentials. This dict has the structure
             {"region": "", "key": "", "secret": "", "sessiontoken": ""}. If credentials could not be resolved,
             this returns None
        """
        result: Dict[str, str] = {}

        # Load the credentials from profile/environment
        creds = self._get_session_creds()

        # After loading credentials, region name might be available here.
        if self._boto3_region:
            result["region"] = self._boto3_region

        if not creds:
            # If we were unable to load credentials, then just return result. We will use the default
            return result

        # Only add the key, if its value is present
        if hasattr(creds, "access_key") and creds.access_key:
            result["key"] = creds.access_key

        if hasattr(creds, "secret_key") and creds.secret_key:
            result["secret"] = creds.secret_key

        if hasattr(creds, "token") and creds.token:
            result["sessiontoken"] = creds.token

        return result



"""
Tarball Archive utility
"""

import os
import tarfile
from contextlib import contextmanager
from tempfile import TemporaryFile
from typing import IO, Optional, Union


@contextmanager
def create_tarball(tar_paths, tar_filter=None, mode="w"):
    """
    Context Manger that creates the tarball of the Docker Context to use for building the image

    Parameters
    ----------
    tar_paths dict(str, str)
        Key representing a full path to the file or directory and the Value representing the path within the tarball

    mode str
        The mode in which the tarfile is opened. Defaults to "w".

    Yields
    ------
    IO
        The tarball file
    """
    tarballfile = TemporaryFile()

    with tarfile.open(fileobj=tarballfile, mode=mode) as archive:
        for path_on_system, path_in_tarball in tar_paths.items():
            archive.add(path_on_system, arcname=path_in_tarball, filter=tar_filter)

    # Flush are seek to the beginning of the file
    tarballfile.flush()
    tarballfile.seek(0)

    try:
        yield tarballfile
    finally:
        tarballfile.close()


def _is_within_directory(directory: Union[str, os.PathLike], target: Union[str, os.PathLike]) -> bool:
    """Checks if target is located under directory"""
    abs_directory = os.path.abspath(directory)
    abs_target = os.path.abspath(target)

    prefix = os.path.commonprefix([abs_directory, abs_target])

    return bool(prefix == abs_directory)


def extract_tarfile(
    tarfile_path: Union[str, os.PathLike] = "",
    file_obj: Optional[IO[bytes]] = None,
    unpack_dir: Union[str, os.PathLike] = "",
) -> None:
    """
    Extracts a tarfile using the provided parameters. If file_obj is specified,
    it is used instead of the file_obj opened for tarfile_path.

    Parameters
    ----------
    tarfile_path Union[str, os.PathLike]
        Key representing a full path to the file or directory and the Value representing the path within the tarball

    file_obj Optional[IO[bytes]]
        Object for the tarfile that will be extracted

    unpack_dir Union[str, os.PathLike]
        The directory where the tarfile members will be extracted.
    """
    with tarfile.open(name=tarfile_path, fileobj=file_obj, mode="r") as tar:
        # Makes sure the tar file is sanitized and is free of directory traversal vulnerability
        # See: https://github.com/advisories/GHSA-gw9q-c7gh-j9vm
        for member in tar.getmembers():
            member_path = os.path.join(unpack_dir, member.name)
            if not _is_within_directory(unpack_dir, member_path):
                raise tarfile.ExtractError("Attempted Path Traversal in Tar File")

        tar.extractall(unpack_dir)



"""
Exceptions that are raised by sam delete
"""
from samcli.commands.exceptions import UserException


class DeleteFailedError(UserException):
    def __init__(self, stack_name, msg, stack_status=None):
        self.stack_name = stack_name
        self.msg = msg
        self.stack_status = stack_status

        message = f"Failed to delete the stack: {stack_name}, msg: {msg}"
        if self.stack_status:
            message += f", status: {self.stack_status}"

        super().__init__(message=message)


class CfDeleteFailedStatusError(UserException):
    def __init__(self, stack_name, msg, stack_status=None):
        self.stack_name = stack_name
        self.msg = msg
        self.stack_status = stack_status

        message = f"Stack {stack_name} could not be deleted as it encountered DELETE_FAILED, " f"msg: {msg}"
        if self.stack_status:
            message += f", status: {self.stack_status}"

        super().__init__(message=message)


class FetchTemplateFailedError(UserException):
    def __init__(self, stack_name, msg):
        self.stack_name = stack_name
        self.msg = msg

        message = f"Failed to fetch the template for the stack: {stack_name}, {msg}"

        super().__init__(message=message)



"""
Wraps watchdog to observe file system for any change.
"""
import logging
import threading
import uuid
from abc import ABC, abstractmethod
from pathlib import Path
from threading import Lock, Thread
from typing import Callable, Dict, List, Optional

import docker
from docker import DockerClient
from docker.errors import ImageNotFound
from docker.types import CancellableStream
from watchdog.events import FileSystemEvent, FileSystemEventHandler, PatternMatchingEventHandler
from watchdog.observers import Observer
from watchdog.observers.api import BaseObserver, ObservedWatch

from samcli.cli.global_config import Singleton
from samcli.lib.utils.hash import dir_checksum, file_checksum
from samcli.lib.utils.packagetype import IMAGE, ZIP
from samcli.local.lambdafn.config import FunctionConfig

LOG = logging.getLogger(__name__)


class ResourceObserver(ABC):
    @abstractmethod
    def watch(self, resource: str) -> None:
        """
        Start watching the input resource.

        Parameters
        ----------
        resource: str
            The resource that should be observed for modifications

        Raises
        ------
        ObserverException:
            if the input resource is not exist
        """

    @abstractmethod
    def unwatch(self, resource: str) -> None:
        """
        Remove the input resource form the observed resorces

        Parameters
        ----------
        resource: str
            The resource to be unobserved
        """

    @abstractmethod
    def start(self):
        """
        Start Observing.
        """

    @abstractmethod
    def stop(self):
        """
        Stop Observing.
        """


class ObserverException(Exception):
    """
    Exception raised when unable to observe the input Lambda Function.
    """


class LambdaFunctionObserver:
    """
    A class that will observe Lambda Function sources regardless if the source is code or image
    """

    def __init__(self, on_change: Callable) -> None:
        """
        Initialize the Image observer
        Parameters
        ----------
        on_change:
            Reference to the function that will be called if there is a change in aby of the observed image
        """
        self._observers: Dict[str, ResourceObserver] = {
            ZIP: FileObserver(self._on_zip_change),
            IMAGE: ImageObserver(self._on_image_change),
        }

        self._observed_functions: Dict[str, Dict[str, List[FunctionConfig]]] = {
            ZIP: {},
            IMAGE: {},
        }

        def _get_zip_lambda_function_paths(function_config: FunctionConfig) -> List[str]:
            """
            Returns a list of ZIP package type lambda function source code paths

            Parameters
            ----------
            function_config: FunctionConfig
                The lambda function configuration that will be observed

            Returns
            -------
            list[str]
                List of lambda functions' source code paths to be observed
            """
            code_paths = [function_config.code_abs_path]
            if function_config.layers:
                # Non-local layers will not have a codeuri property and don't need to be observed
                code_paths += [layer.codeuri for layer in function_config.layers if layer.codeuri]
            return code_paths

        def _get_image_lambda_function_image_names(function_config: FunctionConfig) -> List[str]:
            """
            Returns a list of Image package type lambda function image names

            Parameters
            ----------
            function_config: FunctionConfig
                The lambda function configuration that will be observed

            Returns
            -------
            list[str]
                List of lambda functions' image names to be observed
            """
            return [function_config.imageuri]

        self.get_resources: Dict[str, Callable] = {
            ZIP: _get_zip_lambda_function_paths,
            IMAGE: _get_image_lambda_function_image_names,
        }

        self._input_on_change: Callable = on_change
        self._watch_lock: Lock = threading.Lock()

    def _on_zip_change(self, paths: List[str]) -> None:
        """
        It got executed once there is a change in one of the watched lambda functions' source code.

        Parameters
        ----------
        paths: list[str]
            the changed lambda functions' source code paths
        """
        self._on_change(paths, ZIP)

    def _on_image_change(self, images: List[str]) -> None:
        """
        It got executed once there is a change in one of the watched lambda functions' images.

        Parameters
        ----------
        images: list[str]
            the changed lambda functions' images names
        """
        self._on_change(images, IMAGE)

    def _on_change(self, resources: List[str], package_type: str) -> None:
        """
        It got executed once there is a change in one of the watched lambda functions' resources.

        Parameters
        ----------
        resources: list[str]
            the changed lambda functions' resources (either source code path pr image names)
        package_type: str
            determine if the changed resource is a source code path or an image name
        """
        with self._watch_lock:
            changed_functions: List[FunctionConfig] = []
            for resource in resources:
                if self._observed_functions[package_type].get(resource, None):
                    changed_functions += self._observed_functions[package_type][resource]
            self._input_on_change(changed_functions)

    def watch(self, function_config: FunctionConfig) -> None:
        """
        Start watching the input lambda function.

        Parameters
        ----------
        function_config: FunctionConfig
            The lambda function configuration that will be observed

        Raises
        ------
        ObserverException:
            if not able to observe the input function source path/image
        """
        with self._watch_lock:
            if self.get_resources.get(function_config.packagetype, None):
                resources = self.get_resources[function_config.packagetype](function_config)
                for resource in resources:
                    functions = self._observed_functions[function_config.packagetype].get(resource, [])
                    functions += [function_config]
                    self._observed_functions[function_config.packagetype][resource] = functions
                    self._observers[function_config.packagetype].watch(resource)

    def unwatch(self, function_config: FunctionConfig) -> None:
        """
        Remove the input lambda function from the observed functions

        Parameters
        ----------
        function_config: FunctionConfig
            The lambda function configuration that will be observed
        """
        if self.get_resources.get(function_config.packagetype, None):
            resources = self.get_resources[function_config.packagetype](function_config)
            for resource in resources:
                functions = self._observed_functions[function_config.packagetype].get(resource, [])
                if function_config in functions:
                    functions.remove(function_config)
                if not functions:
                    self._observed_functions[function_config.packagetype].pop(resource, None)
                    self._observers[function_config.packagetype].unwatch(resource)

    def start(self):
        """
        Start Observing.
        """
        for _, observer in self._observers.items():
            observer.start()

    def stop(self):
        """
        Stop Observing.
        """
        for _, observer in self._observers.items():
            observer.stop()


class ImageObserverException(ObserverException):
    """
    Exception raised when unable to observe the input image.
    """


class ImageObserver(ResourceObserver):
    """
    A class that will observe some docker images for any change.
    """

    def __init__(self, on_change: Callable) -> None:
        """
        Initialize the Image observer
        Parameters
        ----------
        on_change:
            Reference to the function that will be called if there is a change in aby of the observed image
        """
        self._observed_images: Dict[str, str] = {}
        self._input_on_change: Callable = on_change
        self.docker_client: DockerClient = docker.from_env()
        self.events: CancellableStream = self.docker_client.events(filters={"type": "image"}, decode=True)
        self._images_observer_thread: Optional[Thread] = None
        self._lock: Lock = threading.Lock()

    def _watch_images_events(self):
        for event in self.events:
            if event.get("Action", None) != "tag":
                continue
            image_name = event["Actor"]["Attributes"]["name"]
            if self._observed_images.get(image_name, None):
                new_image_id = event["id"]
                if new_image_id != self._observed_images[image_name]:
                    self._observed_images[image_name] = new_image_id
                    self._input_on_change([image_name])

    def watch(self, resource: str) -> None:
        """
        Start watching the input image.

        Parameters
        ----------
        resource: str
            The container image name that will be observed

        Raises
        ------
        ImageObserverException:
            if the input image_name is not exist
        """
        try:
            image = self.docker_client.images.get(resource)
            self._observed_images[resource] = image.id
        except ImageNotFound as exc:
            raise ImageObserverException("Can not observe non exist image") from exc

    def unwatch(self, resource: str) -> None:
        """
        Remove the input image form the observed images

        Parameters
        ----------
        resource: str
            The container image name to be unobserved
        """
        self._observed_images.pop(resource, None)

    def start(self):
        """
        Start Observing.
        """
        with self._lock:
            if not self._images_observer_thread:
                self._images_observer_thread = threading.Thread(target=self._watch_images_events, daemon=True)
                self._images_observer_thread.start()

    def stop(self):
        """
        Stop Observing.
        """
        with self._lock:
            self.events.close()
            # wait until the images observer thread got stopped
            while self._images_observer_thread and self._images_observer_thread.is_alive():
                pass


class FileObserverException(ObserverException):
    """
    Exception raised when unable to observe the input path.
    """


class FileObserver(ResourceObserver):
    """
    A class that will Wrap the Singleton File Observer.
    """

    def __init__(self, on_change: Callable) -> None:
        """
        Initialize the file observer
        Parameters
        ----------
        on_change:
            Reference to the function that will be called if there is a change in aby of the observed paths
        """
        self._group = str(uuid.uuid4())
        self._single_file_observer = SingletonFileObserver()
        self._single_file_observer.add_group(self._group, on_change)

    def watch(self, resource: str) -> None:
        self._single_file_observer.watch(resource, self._group)

    def unwatch(self, resource: str) -> None:
        self._single_file_observer.unwatch(resource, self._group)

    def start(self):
        self._single_file_observer.start()

    def stop(self):
        self._single_file_observer.stop()


class SingletonFileObserver(metaclass=Singleton):
    """
    A Singleton class that will observe some file system paths for any change for multiple purposes.
    """

    def __init__(self) -> None:
        """
        Initialize the file observer
        """
        self._observed_paths_per_group: Dict[str, Dict[str, str]] = {}
        self._observed_groups_handlers: Dict[str, Callable] = {}
        self._observed_watches: Dict[str, ObservedWatch] = {}
        self._watch_dog_observed_paths: Dict[str, List[str]] = {}
        self._observer: BaseObserver = Observer()
        self._code_modification_handler: PatternMatchingEventHandler = PatternMatchingEventHandler(
            patterns=["*"], ignore_patterns=[], ignore_directories=False
        )

        self._code_deletion_handler: PatternMatchingEventHandler = PatternMatchingEventHandler(
            patterns=["*"], ignore_patterns=[], ignore_directories=False
        )

        self._code_modification_handler.on_modified = self.on_change
        self._code_deletion_handler.on_deleted = self.on_change
        self._watch_lock = threading.Lock()
        self._lock: Lock = threading.Lock()

    def on_change(self, event: FileSystemEvent) -> None:
        """
        It got executed once there is a change in one of the paths that watchdog is observing.
        This method will check if any of the input paths is really changed, and based on that it will
        invoke the input on_change function with the changed paths

        Parameters
        ----------
        event: watchdog.events.FileSystemEvent
            Determines that there is a change happened to some file/dir in the observed paths
        """
        with self._watch_lock:
            LOG.debug("a %s change got detected in path %s", event.event_type, event.src_path)
            for group, _observed_paths in self._observed_paths_per_group.items():
                if event.event_type == "deleted":
                    observed_paths = [
                        path
                        for path in _observed_paths
                        if path == event.src_path
                        or path in self._watch_dog_observed_paths.get(f"{event.src_path}_False", [])
                    ]
                else:
                    observed_paths = [path for path in _observed_paths if event.src_path.startswith(path)]

                if not observed_paths:
                    continue

                LOG.debug("affected paths of this change %s", observed_paths)
                changed_paths = []
                for path in observed_paths:
                    path_obj = Path(path)
                    # The path got deleted
                    if not path_obj.exists():
                        _observed_paths.pop(path, None)
                        changed_paths += [path]
                    else:
                        new_checksum = calculate_checksum(path)
                        if new_checksum and new_checksum != _observed_paths.get(path, None):
                            changed_paths += [path]
                            _observed_paths[path] = new_checksum
                        else:
                            LOG.debug("the path %s content does not change", path)

                if changed_paths:
                    self._observed_groups_handlers[group](changed_paths)

    def add_group(self, group: str, on_change: Callable) -> None:
        """
        Add new group to file observer. This enable FileObserver to watch the same path for
        multiple purposes.

        Parameters
        ----------
        group: str
            unique string define a new group of paths to be watched.

        on_change: Callable
            The method to be called in case if any path related to this group got changed.
        """
        if group in self._observed_paths_per_group:
            raise Exception(f"The group {group} of paths is already watched")
        self._observed_paths_per_group[group] = {}
        self._observed_groups_handlers[group] = on_change

    def watch(self, resource: str, group: str) -> None:
        """
        Start watching the input path. File Observer will keep track of the input path with its hash, to check it later
        if it got really changed or not.
        File Observer will send the parent path to watchdog for to be observed to avoid the missing events if the input
        paths got deleted.

        Parameters
        ----------
        resource: str
            The file/dir path to be observed

        group: str
            unique string define a new group of paths to be watched.

        Raises
        ------
        FileObserverException:
            if the input path is not exist
        """
        with self._watch_lock:
            path_obj = Path(resource)
            if not path_obj.exists():
                raise FileObserverException("Can not observe non exist path")

            _observed_paths = self._observed_paths_per_group[group]
            _check_sum = calculate_checksum(resource)
            if not _check_sum:
                raise Exception(f"Failed to calculate the hash of resource {resource}")
            _observed_paths[resource] = _check_sum

            LOG.debug("watch resource %s", resource)
            # recursively watch the input path, and all child path for any modification
            self._watch_path(resource, resource, self._code_modification_handler, True)

            LOG.debug("watch resource %s's parent %s", resource, str(path_obj.parent))
            # watch only the direct parent path child directories for any deletion
            # Parent directory watching is needed, as if the input path got deleted,
            # watchdog will not send an event for it
            self._watch_path(str(path_obj.parent), resource, self._code_deletion_handler, False)

    def _watch_path(
        self, watch_dog_path: str, original_path: str, watcher_handler: FileSystemEventHandler, recursive: bool
    ) -> None:
        """
        update the observed paths data structure, and call watch dog observer to observe the input watch dog path
        if it is not observed before

        Parameters
        ----------
        watch_dog_path: str
            The file/dir path to be observed by watch dog
        original_path: str
            The original input file/dir path to be observed
        watcher_handler: FileSystemEventHandler
            The watcher event handler
        recursive: bool
            determines if we need to watch the path, and all children paths recursively, or just the direct children
            paths
        """

        # Allow watching the same path in 2 Modes recursivly, and non-recusrsivly.
        # here, we need to only watch the input path in a specific recursive mode
        original_watch_dog_path = watch_dog_path
        watch_dog_path = f"{watch_dog_path}_{recursive}"
        child_paths = self._watch_dog_observed_paths.get(watch_dog_path, [])
        first_time = not bool(child_paths)
        if original_path not in child_paths:
            child_paths += [original_path]
        self._watch_dog_observed_paths[watch_dog_path] = child_paths
        if first_time:
            LOG.debug("Create Observer for resource %s with recursive %s", original_watch_dog_path, recursive)
            self._observed_watches[watch_dog_path] = self._observer.schedule(
                watcher_handler, original_watch_dog_path, recursive=recursive
            )

    def unwatch(self, resource: str, group: str) -> None:
        """
        Remove the input path form the observed paths, and stop watching this path.

        Parameters
        ----------
        resource: str
            The file/dir path to be unobserved
        group: str
            unique string define a new group of paths to be watched.
        """
        path_obj = Path(resource)

        LOG.debug("unwatch resource %s", resource)
        # unwatch input path
        self._unwatch_path(resource, resource, group, True)

        LOG.debug("unwatch resource %s's parent %s", resource, str(path_obj.parent))
        # unwatch parent path
        self._unwatch_path(str(path_obj.parent), resource, group, False)

    def _unwatch_path(self, watch_dog_path: str, original_path: str, group: str, recursive: bool) -> None:
        """
        update the observed paths data structure, and call watch dog observer to unobserve the input watch dog path
        if it is not observed before

        Parameters
        ----------
        watch_dog_path: str
            The file/dir path to be unobserved by watch dog
        original_path: str
            The original input file/dir path to be unobserved
        group: str
            unique string define a new group of paths to be watched.
        recursive: bool
            determines if we need to watch the path, and all children paths recursively, or just the direct children
            paths
        """

        # Allow watching the same path in 2 Modes recursivly, and non-recusrsivly.
        # here, we need to only stop watching the input path in a specific recursive mode
        original_watch_dog_path = watch_dog_path
        watch_dog_path = f"{watch_dog_path}_{recursive}"
        _observed_paths = self._observed_paths_per_group[group]
        child_paths = self._watch_dog_observed_paths.get(watch_dog_path, [])
        if original_path in child_paths:
            child_paths.remove(original_path)
            _observed_paths.pop(original_path, None)
        if not child_paths:
            self._watch_dog_observed_paths.pop(watch_dog_path, None)
            if self._observed_watches.get(watch_dog_path, None):
                LOG.debug("Unschedule Observer for resource %s with recursive %s", original_watch_dog_path, recursive)
                self._observer.unschedule(self._observed_watches[watch_dog_path])
                self._observed_watches.pop(watch_dog_path, None)

    def start(self):
        """
        Start Observing.
        """
        with self._lock:
            if not self._observer.is_alive():
                self._observer.start()

    def stop(self):
        """
        Stop Observing.
        """
        with self._lock:
            if self._observer.is_alive():
                self._observer.stop()


def calculate_checksum(path: str) -> Optional[str]:
    try:
        path_obj = Path(path)
        if path_obj.is_file():
            checksum = file_checksum(path)
        else:
            checksum = dir_checksum(path)
        return checksum
    except Exception:
        return None



"""
Context information passed to each CLI command
"""

import logging
import uuid
from typing import List, Optional, cast

import click

from samcli.cli.formatters import RootCommandHelpTextFormatter
from samcli.commands.exceptions import CredentialsError
from samcli.lib.utils.sam_logging import (
    LAMBDA_BULDERS_LOGGER_NAME,
    SAM_CLI_FORMATTER_WITH_TIMESTAMP,
    SAM_CLI_LOGGER_NAME,
    SamCliLogger,
)


class Context:
    """
    Top level context object for the CLI. Exposes common functionality required by a CLI, including logging,
    environment config parsing, debug logging etc.

    This object is passed by Click to every command that adds the proper annotation.
    Read this for more details on Click Context - http://click.pocoo.org/5/commands/#nested-handling-and-contexts
    Each command gets its own context object, but linked to both parent and child command's context, like a Linked List.

    This class itself does not rely on how Click works. It is just a plain old Python class that holds common
    properties used by every CLI command.
    """

    _session_id: str
    formatter_class = RootCommandHelpTextFormatter

    def __init__(self):
        """
        Initialize the context with default values
        """
        self._debug = False
        self._aws_region = None
        self._aws_profile = None
        self._session_id = str(uuid.uuid4())
        self._experimental = False
        self._exception = None

    @property
    def exception(self):
        return self._exception

    @exception.setter
    def exception(self, value: Exception):
        """
        Save the exception to handler in the future

        Parameter
        ---------
        value: Exception
            The exception to save for future handling
        """
        self._exception = value

    @property
    def debug(self):
        return self._debug

    @debug.setter
    def debug(self, value):
        """
        Turn on debug logging if necessary.

        :param value: Value of debug flag
        """
        self._debug = value

        if self._debug:
            # Turn on debug logging and display timestamps
            sam_cli_logger = logging.getLogger(SAM_CLI_LOGGER_NAME)
            lambda_builders_logger = logging.getLogger(LAMBDA_BULDERS_LOGGER_NAME)
            SamCliLogger.configure_logger(sam_cli_logger, SAM_CLI_FORMATTER_WITH_TIMESTAMP, logging.DEBUG)
            SamCliLogger.configure_logger(lambda_builders_logger, SAM_CLI_FORMATTER_WITH_TIMESTAMP, logging.DEBUG)

    @property
    def region(self):
        return self._aws_region

    @region.setter
    def region(self, value):
        """
        Set AWS region
        """
        self._aws_region = value
        self._refresh_session()

    @property
    def profile(self):
        return self._aws_profile

    @profile.setter
    def profile(self, value):
        """
        Set AWS profile for credential resolution
        """
        self._aws_profile = value
        self._refresh_session()

    @property
    def session_id(self) -> str:
        """
        Returns the ID of this command session. This is a randomly generated UUIDv4 which will not change until the
        command terminates.
        """
        return self._session_id

    @property
    def experimental(self):
        return self._experimental

    @experimental.setter
    def experimental(self, value):
        self._experimental = value

    @property
    def command_path(self):
        """
        Returns the full path of the command as invoked ex: "sam local generate-event s3 put". Wrapper to
        https://click.palletsprojects.com/en/7.x/api/#click.Context.command_path

        Returns
        -------
        str
            Full path of the command invoked
        """

        # Uses Click's Core Context. Note, this is different from this class, also confusingly named `Context`.
        # Click's Core Context object is the one that contains command path information.
        click_core_ctx = click.get_current_context()
        if click_core_ctx:
            return click_core_ctx.command_path

        return None

    @property
    def template_dict(self):
        """
        Returns the template_dictionary from click context.
        Returns
        -------
        dict
            Template as dictionary

        """
        click_core_ctx = click.get_current_context()
        if click_core_ctx:
            return click_core_ctx.template_dict

        return None

    @staticmethod
    def get_current_context() -> Optional["Context"]:
        """
        Get the current Context object from Click's context stacks. This method is safe to run within the
        actual command's handler that has a ``@pass_context`` annotation. Outside of the handler, you run
        the risk of creating a new Context object which is entirely different from the Context object used by your
        command.
         .. code:
            @pass_context
            def my_command_handler(ctx):
                 # You will get the right context from within the command handler. This will also work from any
                # downstream method invoked as part of the handler.
                 this_context = Context.get_current_context()
                assert ctx == this_context
         Returns
        -------
        samcli.cli.context.Context
            Instance of this object, if we are running in a Click command. None otherwise.
        """

        # Click has the concept of Context stacks. Think of them as linked list containing custom objects that are
        # automatically accessible at different levels. We start from the Core Click context and discover the
        # SAM CLI command-specific Context object which contains values for global options used by all commands.
        #
        # https://click.palletsprojects.com/en/7.x/complex/#ensuring-object-creation
        #

        click_core_ctx = click.get_current_context()
        if click_core_ctx:
            return cast("Context", click_core_ctx.find_object(Context) or click_core_ctx.ensure_object(Context))

        return None

    def _refresh_session(self):
        """
        Update boto3's default session by creating a new session based on values set in the context. Some properties of
        the Boto3's session object are read-only. Therefore when Click parses new AWS session related properties (like
        region & profile), it will call this method to create a new session with latest values for these properties.
        """
        import boto3
        from botocore import credentials, exceptions, session

        try:
            botocore_session = session.get_session()
            boto3.setup_default_session(
                botocore_session=botocore_session, region_name=self._aws_region, profile_name=self._aws_profile
            )
            # get botocore session and setup caching for MFA based credentials
            botocore_session.get_component("credential_provider").get_provider(
                "assume-role"
            ).cache = credentials.JSONFileCache()

        except exceptions.ProfileNotFound as ex:
            raise CredentialsError(str(ex)) from ex


def get_cmd_names(cmd_name, ctx) -> List[str]:
    """
    Given the click core context, return a list representing all the subcommands passed to the CLI

    Parameters
    ----------
    cmd_name : str
        name of current command
    ctx : click.Context
        click context

    Returns
    -------
    list(str)
        List containing subcommand names. Ex: ["local", "start-api"]

    """
    if not ctx:
        return []

    if ctx and not getattr(ctx, "parent", None):
        return [ctx.info_name]
    # Find parent of current context
    _parent = ctx.parent
    _cmd_names = []
    # Need to find the total set of commands that current command is part of.
    if cmd_name != ctx.info_name:
        _cmd_names = [cmd_name]
    _cmd_names.append(ctx.info_name)
    # Go through all parents till a parent of a context exists.
    while _parent.parent:
        info_name = _parent.info_name
        _cmd_names.append(info_name)
        _parent = _parent.parent

    # Make sure the output reads natural. Ex: ["local", "start-api"]
    _cmd_names.reverse()
    return _cmd_names



"""Local Lambda Service that only invokes a function"""

import io
import json
import logging

from flask import Flask, request
from werkzeug.routing import BaseConverter

from samcli.commands.local.lib.exceptions import UnsupportedInlineCodeError
from samcli.lib.utils.stream_writer import StreamWriter
from samcli.local.lambdafn.exceptions import FunctionNotFound
from samcli.local.services.base_local_service import BaseLocalService, LambdaOutputParser

from .lambda_error_responses import LambdaErrorResponses

LOG = logging.getLogger(__name__)


class FunctionNamePathConverter(BaseConverter):
    regex = ".+"
    weight = 300
    part_isolating = False

    def to_python(self, value):
        return value

    def to_url(self, value):
        return value


class LocalLambdaInvokeService(BaseLocalService):
    def __init__(self, lambda_runner, port, host, stderr=None):
        """
        Creates a Local Lambda Service that will only response to invoking a function

        Parameters
        ----------
        lambda_runner samcli.commands.local.lib.local_lambda.LocalLambdaRunner
            The Lambda runner class capable of invoking the function
        port int
            Optional. port for the service to start listening on
        host str
            Optional. host to start the service on
        stderr io.BaseIO
            Optional stream where the stderr from Docker container should be written to
        """
        super().__init__(lambda_runner.is_debugging(), port=port, host=host)
        self.lambda_runner = lambda_runner
        self.stderr = stderr

    def create(self):
        """
        Creates a Flask Application that can be started.
        """
        self._app = Flask(__name__)

        # add converter to support nested stack function path
        self._app.url_map.converters["function_path"] = FunctionNamePathConverter

        path = "/2015-03-31/functions/<function_path:function_name>/invocations"
        self._app.add_url_rule(
            path,
            endpoint=path,
            view_func=self._invoke_request_handler,
            methods=["POST"],
            provide_automatic_options=False,
        )

        # setup request validation before Flask calls the view_func
        self._app.before_request(LocalLambdaInvokeService.validate_request)

        self._construct_error_handling()

    @staticmethod
    def validate_request():
        """
        Validates the incoming request

        The following are invalid
            1. The Request data is not json serializable
            2. Query Parameters are sent to the endpoint
            3. The Request Content-Type is not application/json
            4. 'X-Amz-Log-Type' header is not 'None'
            5. 'X-Amz-Invocation-Type' header is not 'RequestResponse'

        Returns
        -------
        flask.Response
            If the request is not valid a flask Response is returned

        None:
            If the request passes all validation
        """
        flask_request = request
        request_data = flask_request.get_data()

        if not request_data:
            request_data = b"{}"

        request_data = request_data.decode("utf-8")

        try:
            json.loads(request_data)
        except ValueError as json_error:
            LOG.debug("Request body was not json. Exception: %s", str(json_error))
            return LambdaErrorResponses.invalid_request_content(
                "Could not parse request body into json: No JSON object could be decoded"
            )

        if flask_request.args:
            LOG.debug("Query parameters are in the request but not supported")
            return LambdaErrorResponses.invalid_request_content("Query Parameters are not supported")

        request_headers = flask_request.headers

        log_type = request_headers.get("X-Amz-Log-Type", "None")
        if log_type != "None":
            LOG.debug("log-type: %s is not supported. None is only supported.", log_type)
            return LambdaErrorResponses.not_implemented_locally(
                "log-type: {} is not supported. None is only supported.".format(log_type)
            )

        invocation_type = request_headers.get("X-Amz-Invocation-Type", "RequestResponse")
        if invocation_type != "RequestResponse":
            LOG.warning("invocation-type: %s is not supported. RequestResponse is only supported.", invocation_type)
            return LambdaErrorResponses.not_implemented_locally(
                "invocation-type: {} is not supported. RequestResponse is only supported.".format(invocation_type)
            )

        return None

    def _construct_error_handling(self):
        """
        Updates the Flask app with Error Handlers for different Error Codes

        """
        self._app.register_error_handler(500, LambdaErrorResponses.generic_service_exception)
        self._app.register_error_handler(404, LambdaErrorResponses.generic_path_not_found)
        self._app.register_error_handler(405, LambdaErrorResponses.generic_method_not_allowed)

    def _invoke_request_handler(self, function_name):
        """
        Request Handler for the Local Lambda Invoke path. This method is responsible for understanding the incoming
        request and invoking the Local Lambda Function

        Parameters
        ----------
        function_name str
            Name of the function to invoke

        Returns
        -------
        A Flask Response response object as if it was returned from Lambda
        """
        flask_request = request

        request_data = flask_request.get_data()

        if not request_data:
            request_data = b"{}"

        request_data = request_data.decode("utf-8")

        stdout_stream = io.BytesIO()
        stdout_stream_writer = StreamWriter(stdout_stream, auto_flush=True)

        try:
            self.lambda_runner.invoke(function_name, request_data, stdout=stdout_stream_writer, stderr=self.stderr)
        except FunctionNotFound:
            LOG.debug("%s was not found to invoke.", function_name)
            return LambdaErrorResponses.resource_not_found(function_name)
        except UnsupportedInlineCodeError:
            return LambdaErrorResponses.not_implemented_locally(
                "Inline code is not supported for sam local commands. Please write your code in a separate file."
            )

        lambda_response, is_lambda_user_error_response = LambdaOutputParser.get_lambda_output(stdout_stream)

        if is_lambda_user_error_response:
            return self.service_response(
                lambda_response, {"Content-Type": "application/json", "x-amz-function-error": "Unhandled"}, 200
            )

        return self.service_response(lambda_response, {"Content-Type": "application/json"}, 200)
